from __future__ import annotations

from asyncio import CancelledError
from typing import Final, TYPE_CHECKING, Callable, Literal

import HABApp
from HABApp.core.connections._definitions import ConnectionStatus, connection_log
from HABApp.core.connections.status_transitions import StatusTransitions
from HABApp.core.lib import SingleTask, PriorityList
from ..wrapper import process_exception

if TYPE_CHECKING:
    from .base_plugin import BaseConnectionPlugin
    from .plugin_callback import PluginCallbackHandler


class AlreadyHandledException(Exception):
    pass


class HandleExceptionInConnection:
    def __init__(self, connection: BaseConnection, func_name: Callable):
        self._connection: Final = connection
        self._func_name: Final = func_name

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        # no exception -> we exit gracefully
        if exc_type is None and exc_val is None:
            return True

        if isinstance(exc_val, CancelledError):
            return None

        self._connection.process_exception(exc_val, self._func_name)
        raise AlreadyHandledException from None


class BaseConnection:
    def __init__(self, name: str):
        self.name: Final = name
        self.log: Final = connection_log.getChild(name)
        self.status: Final = StatusTransitions()

        # this can be an arbitrary obj that can be reused
        self.context = None

        # Plugin handling
        self.plugins: list[BaseConnectionPlugin] = []
        self.plugin_callbacks: dict[ConnectionStatus, PriorityList[PluginCallbackHandler]] = {
            name: PriorityList() for name in ConnectionStatus}

        # Tasks
        self.plugin_task: Final = SingleTask(self._task_plugin, f'{name.title():s}PluginTask')
        self.advance_status_task: Final = SingleTask(self._task_next_status, f'{name.title():s}AdvanceStatusTask')

    @property
    def is_online(self) -> bool:
        return self.status.status == ConnectionStatus.ONLINE

    @property
    def is_connected(self) -> bool:
        return self.status.status == ConnectionStatus.CONNECTED

    @property
    def is_disconnected(self) -> bool:
        return self.status.status == ConnectionStatus.DISCONNECTED

    @property
    def is_disabled(self) -> bool:
        return self.status.status == ConnectionStatus.DISABLED

    @property
    def is_shutdown(self) -> bool:
        return self.status.status == ConnectionStatus.SHUTDOWN

    @property
    def has_errors(self) -> bool:
        return self.status.error

    def handle_exception(self, func: Callable) -> HandleExceptionInConnection:
        return HandleExceptionInConnection(self, func)

    def is_silent_exception(self, e: Exception):
        return False

    def process_exception(self, e: Exception, func: Callable | str | None):
        self.set_error()

        if self.is_silent_exception(e):
            if func is None:
                self.log.debug(f'Error: {e} ({e.__class__.__name__})')
            else:
                name = func if isinstance(func, str) else func.__qualname__
                self.log.debug(f'Error in {name:s}: {e} ({e.__class__.__name__})')
        else:
            process_exception(func, e, self.log)

    def register_plugin(self, obj: BaseConnectionPlugin, priority: int | Literal['first', 'last'] | None = None):
        from .plugin_callback import get_plugin_callbacks

        # Possibility to specify default priority as a class variable
        if priority is None:
            priority = getattr(obj, '_DEFAULT_PRIORITY', None)

        # Check that it's not already registered
        assert not obj.plugin_callbacks

        for p in self.plugins:
            if p.plugin_name == obj.plugin_name:
                raise ValueError(f'Plugin with the same name already registered: {p}')

        for status, handler in get_plugin_callbacks(obj):
            if priority is None:
                # Handler runs first for every step, except disconnect & offline - there it runs last.
                # That way it's possible to do some cleanup in the plugins when we gracefully disconnect
                if status is ConnectionStatus.DISCONNECTED or status is ConnectionStatus.OFFLINE:
                    self.plugin_callbacks[status].append(handler, 'last')
                else:
                    self.plugin_callbacks[status].append(handler, 'first')
            else:
                self.plugin_callbacks[status].append(handler, priority)

        obj.plugin_connection = self
        self.plugins.append(obj)

        self.log.debug(f'Added plugin {obj.plugin_name:s}')
        return self

    def remove_plugin(self, obj: BaseConnectionPlugin):
        self.plugins.remove(obj)
        obj.plugin_connection = None

        for cb_list in self.plugin_callbacks.values():
            rem = [cb for cb in cb_list if cb.plugin is obj]
            for to_rem in rem:
                cb_list.remove(to_rem)

    async def _task_next_status(self):
        with HABApp.core.wrapper.ExceptionToHABApp(logger=self.log):

            # if we are currently running stop the task
            await self.plugin_task.cancel_wait()

            while (next_value := self.status.advance_status()) is not None:
                self.log.debug(next_value.value)

                assert not self.plugin_task.is_running
                self.plugin_task.start()
                await self.plugin_task.wait()

        self.advance_status_task.task = None

    async def _task_plugin(self):
        status = self.status
        status_enum = status.status

        self.log.debug(f'Task {status_enum.value:s} start')

        callbacks = self.plugin_callbacks[status_enum]
        for cb in callbacks:

            error = True

            try:
                await cb.run(self, self.context)
                error = False
            except AlreadyHandledException:
                pass
            except Exception as e:
                self.process_exception(e, cb.coro)

            if error:
                # Fail fast during connection
                if status.is_connecting_or_connected():
                    break

        self.log.debug(f'Task {status_enum.value:s} done')

    def clear_error(self):
        if not self.status.error:
            return None
        self.log.debug('Cleared error')
        self.status.error = False

    def set_error(self):
        if self.status.error:
            self.log.debug('Error on connection status is already set')
        else:
            self.status.error = True
            self.log.debug('Set error on connection status')
        self.advance_status_task.start_if_not_running()

    def status_from_setup_to_disabled(self):
        self.status.from_setup_to_disabled()
        self.advance_status_task.start_if_not_running()

    def status_from_connected_to_disconnected(self):
        self.status.from_connected_to_disconnected()
        self.advance_status_task.start_if_not_running()

    def status_configuration_changed(self):
        self.log.debug('Requesting setup')
        self.status.setup = True
        self.advance_status_task.start_if_not_running()

    def on_application_shutdown(self):
        if self.status.shutdown:
            return None
        self.log.debug('Requesting shutdown')
        self.status.shutdown = True

        for p in self.plugins:
            p.on_application_shutdown()

        self.advance_status_task.start_if_not_running()

    def application_startup_complete(self):
        self.log.debug('Overview')
        for status, objs in self.plugin_callbacks.items():
            if not objs:
                continue
            coros = []
            for obj in objs:
                name = obj.coro.__name__
                for replace in (f'on_{status.value.lower():s}', f'_on_{status.value.lower():s}', ):
                    if name.startswith(replace):
                        name = name[len(replace):]
                        break
                coros.append(f'{obj.plugin.plugin_name}{"." if name else ""}{name}')

            self.log.debug(f' - {status}: {", ".join(coros)}')

        self.status.setup = True
        self.advance_status_task.start_if_not_running()



import re
from pathlib import Path
from packaging.version import VERSION_PATTERN
import HABApp.__check_dependency_packages__
from HABApp import __version__


def test_installation_check():
    re_name = re.compile(r'^([A-Za-z_-]{3,})')
    requirements = Path(__file__).parent.parent / 'requirements_setup.txt'
    assert requirements.is_file()

    found = set()
    for line in requirements.read_text().splitlines():
        line = line.strip()
        if not line or line.startswith('#'):
            continue
        found.add(re_name.search(line).group(1))

    coded = set(HABApp.__check_dependency_packages__.get_dependencies())

    assert coded == found


def test_version():
    check = re.compile(VERSION_PATTERN, re.VERBOSE | re.IGNORECASE)
    assert check.fullmatch(__version__)



import typing

import paho.mqtt.client as mqtt

import HABApp
from HABApp.mqtt.connection.mqtt_connection import STATUS, log
from ..core.const.json import dump_json


def __is_connected() -> bool:
    if STATUS.connected:
        return True
    raise ConnectionError('Mqtt client not connected')


def publish(topic: str, payload: typing.Any, qos: int = None, retain: bool = None) -> int:
    """
    Publish a value under a certain topic.
    If qos and/or retain is not set the value from the configuration file will be used.

    :param topic: MQTT topic
    :param payload: MQTT Payload
    :param qos: QoS, can be 0, 1 or 2. If not specified value from configuration file will be used.
    :param retain: retain message. If not specified value from configuration file will be used.
    :return: 0 if successful
    """

    assert isinstance(topic, str), type(topic)
    assert isinstance(qos, int) or qos is None, type(qos)
    assert isinstance(retain, bool) or retain is None, type(retain)

    config = HABApp.config.CONFIG.mqtt

    if not __is_connected():
        return mqtt.MQTT_ERR_NO_CONN
    if config.general.listen_only:
        return 100

    if qos is None:
        qos = config.publish.qos
    if retain is None:
        retain = config.publish.retain

    # convert these to string
    if isinstance(payload, (dict, list)):
        payload = dump_json(payload)

    info = STATUS.client.publish(topic, payload, qos, retain)
    if info.rc != mqtt.MQTT_ERR_SUCCESS:
        log.error(f'Could not publish to "{topic}": {mqtt.error_string(info.rc)}')
    return info


def subscribe(topic: str, qos: int = None) -> int:
    """
    Subscribe to a MQTT topic. Note that subscriptions made this way are volatile and will only remain until
    the next disconnect.

    :param topic: MQTT topic to subscribe to
    :param qos: QoS, can be 0, 1 or 2.  If not specified value from configuration file will be used.
    :return: 0 if successful
    """

    assert isinstance(topic, str), type(topic)
    assert isinstance(qos, int) or qos is None, type(qos)

    if not __is_connected():
        return mqtt.MQTT_ERR_NO_CONN

    # If no qos is specified load it from config
    if qos is None:
        qos = HABApp.config.CONFIG.mqtt.subscribe.qos

    res, mid = STATUS.client.subscribe(topic, qos)
    if res != mqtt.MQTT_ERR_SUCCESS:
        log.error(f'Could not subscribe to "{topic}": {mqtt.error_string(res)}')
    return res


def unsubscribe(topic: str) -> int:
    """
    Unsubscribe from a MQTT topic

    :param topic: MQTT topic
    :return: 0 if successful
    """

    assert isinstance(topic, str), type(topic)

    if not __is_connected():
        return mqtt.MQTT_ERR_NO_CONN

    result, mid = STATUS.client.unsubscribe(topic)
    if result != mqtt.MQTT_ERR_SUCCESS:
        log.error(f'Could not unsubscribe from "{topic}": {mqtt.error_string(result)}')
    return result



from . import habapp_events
from .events import ComplexEventValue, ValueUpdateEvent, ValueChangeEvent, \
    ItemNoChangeEvent, ItemNoUpdateEvent
from .filter import NoEventFilter, OrFilterGroup, AndFilterGroup, ValueUpdateEventFilter, ValueChangeEventFilter, \
    EventFilter



from typing import TYPE_CHECKING, Tuple, Optional, FrozenSet, Mapping

from HABApp.openhab.definitions import OnOffValue
from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from HABApp.openhab.items.commands import OnOffCommand

if TYPE_CHECKING:
    Tuple = Tuple
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


ON = OnOffValue.ON
OFF = OnOffValue.OFF


class SwitchItem(OpenhabItem, OnOffCommand):
    """SwitchItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """


    @staticmethod
    def _state_from_oh_str(state: str):
        if state != ON and state != OFF:
            raise ValueError(f'Invalid value for SwitchItem: {state}')
        return state

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OnOffValue):
            new_value = new_value.value

        if new_value is not None and new_value != ON and new_value != OFF:
            raise ValueError(f'Invalid value for SwitchItem {self.name}: {new_value}')
        return super().set_value(new_value)

    def is_on(self) -> bool:
        """Test value against on-value"""
        return self.value == ON

    def is_off(self) -> bool:
        """Test value against off-value"""
        return self.value == OFF

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, SwitchItem):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, int):
            if other and self.is_on():
                return True
            if not other and self.is_off():
                return True
            return False

        return NotImplemented

    def __bool__(self):
        return self.is_on()



from __future__ import annotations

import logging
from asyncio import sleep
from time import monotonic

from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.items import OpenhabItem
from HABApp.runtime import shutdown

log = logging.getLogger('HABApp.openhab.startup')

item_registry = uses_item_registry()


def count_none_items() -> int:
    found = 0
    for item in item_registry.get_items():
        if isinstance(item, OpenhabItem) and item.value is None:
            found += 1
    return found


class WaitForPersistenceRestore(BaseConnectionPlugin[OpenhabConnection]):

    async def on_connected(self, context: OpenhabContext):
        if context.waited_for_openhab:
            log.debug('Openhab has already been running -> complete')
            return None

        # if we find None items check if they are still getting initialized (e.g. from persistence)
        if this_count := count_none_items():
            log.debug('Some items are still None - waiting for initialisation')

            last_count = -1
            start = monotonic()

            while not shutdown.requested and last_count != this_count:
                await sleep(2)

                # timeout so we start eventually
                if monotonic() - start >= 180:
                    log.debug('Timeout while waiting for initialisation')
                    break

                last_count = this_count
                this_count = count_none_items()

        log.debug('complete')



from __future__ import annotations

import logging
from typing import Final, Callable, Awaitable, Any

from HABApp.core.lib import SingleTask


class PluginTask(SingleTask):
    def __init__(self, coro: Callable[[], Awaitable[Any]], name: str | None,
                 logger: logging.Logger | None,
                 exception_handler: Callable[[Exception, Callable | str | None], Any]):
        super().__init__(coro, name)

        self.log: Final = logger
        self.exception_handler: Final = exception_handler

    async def _task_wrap(self):
        if self.log is not None:
            self.log.debug(f'Task {self.name} start')

        suffix = ''

        try:
            await super()._task_wrap()
        except Exception as e:
            suffix = ' (with error)'
            self.exception_handler(e, self.name)
        finally:
            if self.log is not None:
                self.log.debug(f'Task {self.name} done {suffix}')



from typing import NoReturn

from HABApp.openhab.errors import MapTransformationError


class MapTransformation(dict):
    def __init__(self, *args, name: str) -> None:
        super().__init__(*args)
        self._name = name

    def __repr__(self, additional: str = ''):
        return f'<{self.__class__.__name__} name={self._name} items={super().__repr__()}{additional:s}>'


class MapTransformationWithDefault(MapTransformation):
    def __init__(self, *args, default, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._default = default

    def __missing__(self, key):
        return self._default

    def __repr__(self, additional: str = ''):
        return super().__repr__(f', default={self._default}{additional}')

    def get(self, key, default=None) -> NoReturn:
        raise MapTransformationError(f'Mapping is already defined with a default: "{self._default}"')



from pathlib import Path
from string import Template
from time import sleep

from HABApp.config.platform_defaults import get_log_folder, is_openhabian


def get_default_logfile() -> str:
    template = Template("""
${LOG_LEVELS}formatters:
  HABApp_format:
    format: '[%(asctime)s] [%(name)25s] %(levelname)8s | %(message)s'
${FRONTAIL_FORMAT}

handlers:
  # There are several Handlers available:
  #  - logging.handlers.RotatingFileHandler:
  #    Will rotate when the file reaches a certain size (see python logging documentation for args)
  #  - HABApp.config.logging.MidnightRotatingFileHandler:
  #    Will wait until the file reaches a certain size and then will rotate on midnight
  #  - More handlers:
  #    https://docs.python.org/3/library/logging.handlers.html#rotatingfilehandler

  HABApp_default:
    class: HABApp.config.logging.MidnightRotatingFileHandler
    filename: '${HABAPP_FILE}'
    maxBytes: 1_048_576
    backupCount: 3

    formatter: ${HABAPP_FILE_FORMAT}
    level: DEBUG

  EventFile:
    class: HABApp.config.logging.MidnightRotatingFileHandler
    filename: '${EVENT_FILE}'
    maxBytes: 1_048_576
    backupCount: 3

    formatter: HABApp_format
    level: DEBUG


loggers:
  HABApp:
    level: INFO
    handlers:
      - HABApp_default
    propagate: False

  HABApp.EventBus:
    level: INFO
    handlers:
      - EventFile
    propagate: False

""")

    # Default values are relative
    subs = {
        'EVENT_FILE': 'events.log',
        'HABAPP_FILE': 'HABApp.log',
        'HABAPP_FILE_FORMAT': 'HABApp_format',
        'FRONTAIL_FORMAT': '',
        'LOG_LEVELS': '',
    }

    # Use abs path and rename events.log if we log in the openHAB folder
    log_folder = get_log_folder()
    if log_folder is not None:
        # Absolute so we can log errors if the config is faulty
        subs['HABAPP_FILE'] = (log_folder / subs['HABAPP_FILE']).as_posix()

        # Keep this relative so it is easier to read in the file
        subs['EVENT_FILE'] = 'HABApp_events.log'

        # With openhabian we typically use frontail
        if is_openhabian():
            subs['FRONTAIL_FORMAT'] = '\n'\
                '  Frontail_format:\n' \
                "    format: '%(asctime)s.%(msecs)03d [%(levelname)-5s] [%(name)-36s] - %(message)s'\n" \
                "    datefmt: '%Y-%m-%d %H:%M:%S'"

            subs['HABAPP_FILE_FORMAT'] = 'Frontail_format'

            # frontail expects WARN instead of WARNING
            subs['LOG_LEVELS'] = 'levels:\n  WARNING: WARN\n\n'

    return template.substitute(**subs)


def create_default_logfile(path: Path) -> bool:
    if path.is_file():
        return False

    print(f'Creating {path.name} in {path.parent}')
    with path.open('w', encoding='utf-8') as file:
        file.write(get_default_logfile())

    sleep(0.01)
    return True



from datetime import datetime
from typing import Optional, Union

from HABApp.openhab.definitions.rest import ItemHistoryResp

OPTIONAL_DT = Optional[datetime]


class OpenhabPersistenceData:

    def __init__(self):
        self.data: dict[float, Union[int, float, str]] = {}

    @classmethod
    def from_resp(cls, data: ItemHistoryResp) -> 'OpenhabPersistenceData':
        c = cls()
        for entry in data.data:
            # calc as timestamp
            time = entry.time / 1000
            state = entry.state
            if '.' in state:
                try:
                    state = float(state)
                except ValueError:
                    pass
            else:
                try:
                    state = int(state)
                except ValueError:
                    pass

            c.data[time] = state
        return c

    def get_data(self, start_date: OPTIONAL_DT = None, end_date: OPTIONAL_DT = None):
        if start_date is None and end_date is None:
            return self.data

        filter_start = start_date.timestamp() if start_date else None
        filter_end = end_date.timestamp() if end_date else None

        ret = {}
        for ts, val in self.data.items():
            if filter_start is not None and ts < filter_start:
                continue
            if filter_end is not None and ts > filter_end:
                continue
            ret[ts] = val
        return ret

    def min(self, start_date: OPTIONAL_DT = None, end_date: OPTIONAL_DT = None) -> Optional[float]:
        return min(self.get_data(start_date, end_date).values(), default=None)

    def max(self, start_date: OPTIONAL_DT = None, end_date: OPTIONAL_DT = None) -> Optional[float]:
        return max(self.get_data(start_date, end_date).values(), default=None)

    def average(self, start_date: OPTIONAL_DT = None, end_date: OPTIONAL_DT = None) -> Optional[float]:
        values = list(self.get_data(start_date, end_date).values())
        ct = len(values)
        if ct == 0:
            return None
        return sum(values) / ct



from typing import Optional

from HABApp.core.errors import ItemNotFoundException
from HABApp.core.internals import uses_item_registry
from HABApp.mqtt.interface_sync import publish
from . import MqttBaseItem

Items = uses_item_registry()


def build_write_topic(read_topic: str) -> Optional[str]:
    parts = read_topic.split('/')
    if parts[0] == 'zigbee2mqtt':
        parts.insert(-1, 'set')
        return '/'.join(parts)

    raise ValueError(f'Can not build write topic for "{read_topic}"')


class MqttPairItem(MqttBaseItem):
    """An item that represents both a topic that is used to read
    and a corresponding topic that is used to write values"""

    @classmethod
    def get_create_item(cls, name: str, write_topic: Optional[str] = None, initial_value=None) -> 'MqttPairItem':
        """Creates a new item in HABApp and returns it or returns the already existing one with the given name.
        HABApp tries to automatically derive the write topic from the item name. In cases where this does not
        work it can be specified manually.

        :param name: item name (topic that reports the state)
        :param write_topic: topic that is used to write values or ``None`` (default) to build it automatically
        :param initial_value: state the item will have if it gets created
        :return: item
        """
        assert isinstance(name, str), type(name)

        # try to build write topic
        if write_topic is None:
            write_topic = build_write_topic(name)

        try:
            item = Items.get_item(name)
        except ItemNotFoundException:
            item = Items.add_item(cls(name, write_topic=write_topic, initial_value=initial_value))

        assert isinstance(item, cls), f'{cls} != {type(item)}'
        return item

    def __init__(self, name: str, initial_value=None, write_topic: str = None):
        super().__init__(name, initial_value)
        self.write_topic: str = write_topic

    def publish(self, payload, qos: int = None, retain: bool = None):
        """
        Publish the payload under the write topic from the item.

        :param payload: MQTT Payload
        :param qos: QoS, can be ``0``, ``1`` or ``2``. If not specified value from configuration file will be used.
        :param retain: retain message. If not specified value from configuration file will be used.
        :return: 0 if successful
        """

        return publish(self.write_topic, payload, qos=qos, retain=retain)



from typing import TYPE_CHECKING, Optional, FrozenSet, Mapping, Tuple, Any

from HABApp.core.events import ComplexEventValue
from HABApp.openhab.item_to_reg import get_members
from HABApp.openhab.items.base_item import OpenhabItem, MetaData

if TYPE_CHECKING:
    Any = Any
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


class GroupItem(OpenhabItem):
    """GroupItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Any value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        return state

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, ComplexEventValue):
            new_value = new_value.value
        return super().set_value(new_value)

    @property
    def members(self) -> Tuple[OpenhabItem, ...]:
        """Resolves and then returns all group members"""

        return get_members(self.name)



from typing import Any, TYPE_CHECKING, Optional, FrozenSet, Mapping

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from ..definitions import OpenClosedValue
from ...core.const import MISSING
from ..errors import SendCommandNotSupported
from HABApp.openhab.interface_sync import post_update

if TYPE_CHECKING:
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


OPEN = OpenClosedValue.OPEN
CLOSED = OpenClosedValue.CLOSED


class ContactItem(OpenhabItem):
    """ContactItem

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        if state != OPEN and state != CLOSED:
            raise ValueError(f'Invalid value for ContactItem: {state}')
        return state

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, OpenClosedValue):
            new_value = new_value.value

        if new_value is not None and new_value != OPEN and new_value != CLOSED:
            raise ValueError(f'Invalid value for ContactItem: {new_value}')
        return super().set_value(new_value)

    def is_open(self) -> bool:
        """Test value against open value"""
        return self.value == OPEN

    def is_closed(self) -> bool:
        """Test value against closed value"""
        return self.value == CLOSED

    def oh_send_command(self, value: Any = MISSING):
        raise SendCommandNotSupported(f'{self.__class__.__name__} does not support send command! '
                                      'See openHAB documentation for details.')

    def open(self):
        """Post an update to the item with the open value"""
        return post_update(self.name, OPEN)

    def closed(self):
        """Post an update to the item with the closed value"""
        return post_update(self.name, CLOSED)

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, ContactItem):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        elif isinstance(other, int):
            if other and self.is_open():
                return True
            if not other and self.is_closed():
                return True
            return False

        return NotImplemented



from .items import ShortItemResp, ItemResp
from .things import ThingResp, ChannelResp
from .links import ItemChannelLinkResp

from .root import RootResp
from .systeminfo import SystemInfoRootResp
from .transformations import TransformationResp
from .persistence import ItemHistoryResp, PersistenceServiceResp



from __future__ import annotations

import logging
from typing import Final

from HABApp.core.const.const import StrEnum

connection_log = logging.getLogger('HABApp.connection')


class ConnectionStatus(StrEnum):
    STARTUP = 'STARTUP'

    SETUP = 'SETUP'

    # connection flow
    CONNECTING = 'CONNECTING'
    CONNECTED = 'CONNECTED'
    ONLINE = 'ONLINE'

    # unexpected disconnect or error
    DISCONNECTED = 'DISCONNECTED'
    OFFLINE = 'OFFLINE'

    # connection is disabled
    DISABLED = 'DISABLED'

    # normal shutdown flow
    SHUTDOWN = 'SHUTDOWN'


# Handler which manages the connection
CONNECTION_HANDLER_NAME: Final = 'ConnectionHandler'



from typing import Dict

from msgspec import Struct


# Documentation of TransformationDTO
# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.transform/src/main/java/org/openhab/core/io/rest/transform/TransformationDTO.java

class TransformationResp(Struct):
    uid: str
    label: str
    type: str
    configuration: Dict[str, str]
    editable: bool



from typing import Callable

from HABApp.core.internals.wrapped_function import WrappedFunctionBase
from eascheduler.executors import ExecutorBase


class WrappedFunctionExecutor(ExecutorBase):
    def __init__(self, func: Callable, *args, **kwargs):
        assert isinstance(func, WrappedFunctionBase), type(func)
        super().__init__(func, *args, **kwargs)

    def execute(self):
        self._func.run(*self._args, **self._kwargs)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Any, TYPE_CHECKING

import aiohttp

import HABApp
from HABApp.core.connections import BaseConnection, Connections, ConnectionStateToEventBusPlugin, AutoReconnectPlugin
from HABApp.core.const.const import PYTHON_310
from HABApp.core.items.base_valueitem import datetime

if PYTHON_310:
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from HABApp.openhab.items import OpenhabItem, Thing


@dataclass
class OpenhabContext:
    version: tuple[int, int, int]
    is_oh3: bool

    # true when we waited during connect
    waited_for_openhab: bool

    created_items: dict[str, tuple[OpenhabItem, datetime]]
    created_things: dict[str, tuple[Thing, datetime]]

    session: aiohttp.ClientSession
    session_options: dict[str, Any]


CONTEXT_TYPE: TypeAlias = Optional[OpenhabContext]


def setup():
    config = HABApp.config.CONFIG.openhab

    from HABApp.openhab.connection.handler import HANDLER as CONNECTION_HANDLER
    from HABApp.openhab.connection.plugins import (WaitForStartlevelPlugin, LoadOpenhabItemsPlugin,
                                                   SseEventListenerPlugin, OUTGOING_PLUGIN, LoadTransformationsPlugin,
                                                   WaitForPersistenceRestore, PingPlugin, ThingOverviewPlugin,
                                                   TextualThingConfigPlugin, BrokenLinksPlugin)

    connection = Connections.add(OpenhabConnection())
    connection.register_plugin(CONNECTION_HANDLER)

    connection.register_plugin(WaitForStartlevelPlugin(), 0)
    connection.register_plugin(OUTGOING_PLUGIN, 10)
    connection.register_plugin(LoadOpenhabItemsPlugin('LoadItemsAndThings'), 20)
    connection.register_plugin(SseEventListenerPlugin(), 30)
    connection.register_plugin(LoadOpenhabItemsPlugin('SyncItemsAndThings'), 40)
    connection.register_plugin(LoadTransformationsPlugin(), 50)
    connection.register_plugin(PingPlugin(), 100)
    connection.register_plugin(WaitForPersistenceRestore(), 110)
    connection.register_plugin(TextualThingConfigPlugin(), 120)
    connection.register_plugin(ThingOverviewPlugin(), 500_000)
    connection.register_plugin(BrokenLinksPlugin(), 500_001)

    connection.register_plugin(ConnectionStateToEventBusPlugin())
    connection.register_plugin(AutoReconnectPlugin())

    # config changes
    config.general.subscribe_for_changes(CONNECTION_HANDLER.update_cfg_general)


class OpenhabConnection(BaseConnection):
    def __init__(self):
        super().__init__('openhab')
        self.context: CONTEXT_TYPE = None

    def is_silent_exception(self, e: Exception):
        return isinstance(e, (
            # aiohttp Exceptions
            aiohttp.ClientPayloadError, aiohttp.ClientConnectorError, aiohttp.ClientOSError,

            # aiohttp_sse_client Exceptions
            ConnectionRefusedError, ConnectionError, ConnectionAbortedError)
        )



from __future__ import annotations

from HABApp.core.connections import BaseConnection, BaseConnectionPlugin
from HABApp.core.const.topics import TOPIC_CONNECTIONS
from HABApp.core.events.habapp_events import HABAppConnectionStateEvent
from HABApp.core.internals import uses_post_event


post_event = uses_post_event()


class ConnectionStateToEventBusPlugin(BaseConnectionPlugin):
    _DEFAULT_PRIORITY = 100_000

    def __init__(self, name: str | None = None):
        super().__init__(name)
        self.__last_report = None

    def __post_event(self, connection: BaseConnection):
        if (status := connection.status.status.value) != self.__last_report:
            post_event(TOPIC_CONNECTIONS, HABAppConnectionStateEvent(connection.name, status))
            self.__last_report = status

    async def on_online(self, connection: BaseConnection):
        self.__post_event(connection)

    async def on_disconnected(self, connection: BaseConnection):
        self.__post_event(connection)



from HABApp.core.const.const import StrEnum
from typing import Final


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/ThingStatus.java
class ThingStatusEnum(StrEnum):
    UNINITIALIZED = 'UNINITIALIZED'
    INITIALIZING = 'INITIALIZING'
    UNKNOWN = 'UNKNOWN'
    ONLINE = 'ONLINE'
    OFFLINE = 'OFFLINE'
    REMOVING = 'REMOVING'
    REMOVED = 'REMOVED'


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/ThingStatusDetail.java
class ThingStatusDetailEnum(StrEnum):
    NONE = 'NONE'
    NOT_YET_READY = 'NOT_YET_READY'
    HANDLER_MISSING_ERROR = 'HANDLER_MISSING_ERROR'
    HANDLER_REGISTERING_ERROR = 'HANDLER_REGISTERING_ERROR'
    HANDLER_INITIALIZING_ERROR = 'HANDLER_INITIALIZING_ERROR'
    HANDLER_CONFIGURATION_PENDING = 'HANDLER_CONFIGURATION_PENDING'
    CONFIGURATION_PENDING = 'CONFIGURATION_PENDING'
    COMMUNICATION_ERROR = 'COMMUNICATION_ERROR'
    CONFIGURATION_ERROR = 'CONFIGURATION_ERROR'
    BRIDGE_OFFLINE = 'BRIDGE_OFFLINE'
    FIRMWARE_UPDATING = 'FIRMWARE_UPDATING'
    DUTY_CYCLE = 'DUTY_CYCLE'
    BRIDGE_UNINITIALIZED = 'BRIDGE_UNINITIALIZED'
    GONE = 'GONE'
    DISABLED = 'DISABLED'


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/internal/ThingImpl.java#L67
THING_STATUS_DEFAULT: Final = ThingStatusEnum.UNINITIALIZED
THING_STATUS_DETAIL_DEFAULT: Final = ThingStatusDetailEnum.NONE



import asyncio
import logging
import os
from pathlib import Path
from typing import Optional, Union, Iterable, Any, Tuple, Dict, List, Callable

from typing_extensions import TypeAlias

import HABApp
from HABApp.core.logger import HABAppError, HABAppWarning
from HABApp.core.wrapper import process_exception

log = logging.getLogger('HABApp.execute')

HINT_EXEC_ARGS: TypeAlias = Union[str, Path]
HINT_PYTHON_PATH: TypeAlias = Optional[Iterable[Union[str, Path]]]


def _ensure_str_objs(objs: Iterable[HINT_EXEC_ARGS], key: str, enforce_abs=False) -> List[str]:
    new_args: List[str] = []

    # args must be str, but we support str and Path
    for i, val in enumerate(objs):
        if isinstance(val, Path):
            path_val = val
            str_val = str(val)
        elif isinstance(val, str):
            path_val = Path(val)
            str_val = val
        else:
            raise ValueError(f'{key:s}[{i:d}] is not of type str! "{val}" ({type(val).__name__:s})')

        if enforce_abs and not path_val.is_absolute():
            raise ValueError(f'{key:s}[{i:d}] is not an absolute path: "{val}"')

        new_args.append(str_val)

    return new_args


def build_exec_params(*args: HINT_EXEC_ARGS,
                      _capture_output=True,
                      _additional_python_path: HINT_PYTHON_PATH = None,
                      **kwargs: Any) -> Tuple[Iterable[str], Dict[str, Any]]:

    # convenience for easy capturing
    if _capture_output:
        if 'stdout' in kwargs:
            raise ValueError('Parameter "capture_output" can not be used with "stdout" in kwargs!')
        kwargs['stdout'] = asyncio.subprocess.PIPE
        if 'stderr' in kwargs:
            raise ValueError('Parameter "capture_output" can not be used with "stderr" in kwargs!')
        kwargs['stderr'] = asyncio.subprocess.PIPE

    # convenience for additional libraries
    if _additional_python_path is not None:
        if 'env' in kwargs:
            raise ValueError('Parameter "additional_python_path" can not be used with "env" in kwargs!')

        ppath = _ensure_str_objs(_additional_python_path, 'additional_python_path', enforce_abs=True)

        # The child process needs the env from the current process, that's why we merge them
        # See docs of subprocess.Popen -> env for more details
        env = dict(os.environ)

        existing_ppath = env.get('PYTHONPATH')
        if existing_ppath:
            ppath.append(existing_ppath)
        env['PYTHONPATH'] = os.pathsep.join(ppath)
        kwargs['env'] = env

    # set config folder as working directory
    if 'cwd' not in kwargs:
        kwargs['cwd'] = HABApp.CONFIG._file_path.parent

    return _ensure_str_objs(args, 'args'), kwargs


class FinishedProcessInfo:
    """Information about the finished process.

    :var int returncode: Return code of the process
    :var Optional[str] stdout: Standard output of the process or ``None``
    :var Optional[str] stderr: Error output of the process or ``None``
    """

    def __init__(self, returncode: int, stdout: Optional[str], stderr: Optional[str]):
        self.returncode: int = returncode
        self.stdout: Optional[str] = stdout
        self.stderr: Optional[str] = stderr

    def __repr__(self):
        return f'<ProcessInfo: returncode: {self.returncode}, stdout: {self.stdout}, stderr: {self.stderr}>'

    def __eq__(self, other):
        if isinstance(other, FinishedProcessInfo):
            return self.returncode == other.returncode and self.stdout == other.stdout and self.stderr == other.stderr
        else:
            return NotImplementedError


HINT_PROCESS_CB_FULL: TypeAlias = Callable[[FinishedProcessInfo], Any]
HINT_PROCESS_CB_SIMPLE: TypeAlias = Callable[[str], Any]


async def async_subprocess_exec(callback, *args, calling_func, raw_info: bool, **kwargs):

    call_str = ''

    try:

        proc = None
        stdout = None
        stderr = None

        call_str = ' '.join(map(lambda x: f'"{x}"', args))

        try:
            proc = await asyncio.create_subprocess_exec(*args, **kwargs)

            b_stdout, b_stderr = await proc.communicate()
            ret_code = proc.returncode

            if b_stdout or b_stderr:
                stdout = b_stdout.decode()
                stderr = b_stderr.decode()

        except asyncio.CancelledError:
            if proc is not None:
                proc.terminate()

            HABAppWarning(log=log).add(f'Subprocess canceled! Call: {call_str}').dump()
            return None

    except FileNotFoundError as e:
        HABAppError(log=log).add('Creating subprocess failed!').add(f'  Call: {call_str}').add(
            f'  Working dir: {kwargs.get("cwd")}').add(str(e)).dump()
        return None
    except Exception as e:
        HABAppError(log=log).add('Creating subprocess failed!').add(f'  Call: {call_str}').add(
            f'  Working dir: {kwargs.get("cwd")}').dump()
        process_exception(calling_func, e, logger=log)
        return None

    if raw_info:
        # callback is a wrapped function, that's why it will not throw an error
        callback(FinishedProcessInfo(ret_code, stdout, stderr))
        return None

    if ret_code != 0:
        e = HABAppError(log=log).add(f'Process returned {ret_code:d}!').add(f'Call: {call_str}')
        if stderr:
            e.add('Stderr:')
            for line in stderr.splitlines():
                e.add(f'  {line}')
        elif stdout:
            e.add('Stdout:')
            for line in stdout.splitlines():
                e.add(f'  {line}')
        e.dump()
        return None

    # no error -> callback always with str
    callback(stdout if stdout else '')
    return None



# noinspection PyProtectedMember
from sys import _getframe as sys_get_frame
from typing import Dict, List, Optional, Final, Callable

from HABApp.core.errors import ProxyObjHasNotBeenReplacedError

PROXIES: List['StartUpProxyObj'] = []


class ProxyObjBase:
    @property
    def to_replace_name(self) -> str:
        raise NotImplementedError()

    def __getattr__(self, item):
        raise ProxyObjHasNotBeenReplacedError(self)

    def __call__(self, *args, **kwargs):
        raise ProxyObjHasNotBeenReplacedError(self)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.to_replace_name}>'


class ConstProxyObj(ProxyObjBase):
    def __init__(self, name: str):
        self.name: Final = name

    @property
    def to_replace_name(self) -> str:
        return self.name


class StartUpProxyObj(ProxyObjBase):
    def __init__(self, to_replace: Callable, globals: dict):
        self.to_replace: Optional[Callable] = to_replace
        self.globals: Optional[dict] = globals

        PROXIES.append(self)

    @property
    def to_replace_name(self) -> str:
        return str(getattr(self.to_replace, "__name__", self.to_replace))

    def replace(self, replacements: Dict[object, object], final: bool):
        assert self.globals is not None
        replacement = replacements[self.to_replace]

        for name, value in self.globals.items():
            if value is self:
                self.globals[name] = replacement
                if not final:
                    return RestoreableObj(name, self.globals, self)
                break
        else:
            file = self.globals.get('__file__', '?')
            raise ValueError(f'"{self.to_replace_name}" should be replaced but was not found in {file}!')

        self.globals = None
        self.to_replace = None


def create_proxy(to_replace: Callable) -> StartUpProxyObj:
    frm = sys_get_frame(2)
    return StartUpProxyObj(to_replace, frm.f_globals)


class RestoreableObj:
    def __init__(self, key: str, globals: dict, proxy: 'StartUpProxyObj'):
        self.key = key
        self.globals = globals
        self.proxy = proxy

    def restore(self):
        self.globals[self.key] = self.proxy
        self.globals = None
        self.key = None
        self.proxy = None


def replace_proxies(replacements: Dict[object, object], final: bool) -> List[RestoreableObj]:
    restore_objs = []
    for proxy in PROXIES:
        restore = proxy.replace(replacements, final)
        if restore is not None:
            restore_objs.append(restore)

    if not final:
        return restore_objs

    PROXIES.clear()
    return []



import logging
from asyncio import iscoroutinefunction
from typing import Union, Optional, Callable, Type

from HABApp.config import CONFIG
from HABApp.core.internals import Context
from HABApp.core.internals.wrapped_function.base import TYPE_WRAPPED_FUNC_OBJ
from HABApp.core.internals.wrapped_function.wrapped_async import WrappedAsyncFunction, HINT_FUNC_ASYNC
from HABApp.core.internals.wrapped_function.wrapped_sync import WrappedSyncFunction
from HABApp.core.internals.wrapped_function.wrapped_thread import HINT_FUNC_SYNC, WrappedThreadFunction, \
    create_thread_pool, stop_thread_pool, run_in_thread_pool


def wrap_func(func: Union[HINT_FUNC_SYNC, HINT_FUNC_ASYNC],
              warn_too_long=True,
              name: Optional[str] = None,
              logger: Optional[logging.Logger] = None,
              context: Optional[Context] = None) -> TYPE_WRAPPED_FUNC_OBJ:

    # Check that it's actually a callable, so we fail fast and not when we try to run the function.
    # Some users pass the result of the function call (e.g. func()) by accident
    # which will inevitably fail once we try to run the function.
    if not callable(func):
        try:
            type_name: str = func.__class__.__name__
        except Exception:
            type_name = type(func)
        raise ValueError(f'Callable or coroutine function expected! Got "{func}" (type {type_name:s})')

    if iscoroutinefunction(func):
        return WrappedAsyncFunction(func, name=name, logger=logger, context=context)
    else:
        return SYNC_CLS(func, warn_too_long=warn_too_long, name=name, logger=logger, context=context)


SYNC_CLS: Union[Type[WrappedThreadFunction], Type[WrappedSyncFunction]]


def setup():
    global SYNC_CLS

    if not THREAD_POOL.enabled:
        SYNC_CLS = WrappedSyncFunction

        # In case of hot reload
        stop_thread_pool()
    else:
        SYNC_CLS = WrappedThreadFunction

        # create thread pool
        create_thread_pool(THREAD_POOL.threads)

        # this function can be called multiple times, so it's no problem if we register it more than once!
        from HABApp.runtime import shutdown
        shutdown.register_func(stop_thread_pool, msg='Stopping thread pool', last=True)


THREAD_POOL = CONFIG.habapp.thread_pool
THREAD_POOL.subscribe_for_changes(setup)


async def run_function(func: Callable):
    if not THREAD_POOL.enabled:
        return func()
    else:
        return await run_in_thread_pool(func)



from typing import Any, Dict

from msgspec import Struct, field


class ItemChannelLinkResp(Struct, kw_only=True):
    # AbstractLinkDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/link/dto/AbstractLinkDTO.java
    item: str = field(name='itemName')

    # ItemChannelLinkDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.thing/src/main/java/org/openhab/core/thing/link/dto/ItemChannelLinkDTO.java
    channel: str = field(name='channelUID')
    configuration: Dict[str, Any] = {}

    # EnrichedItemChannelLinkDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/link/EnrichedItemChannelLinkDTO.java
    editable: bool



from typing import List, Dict, Any, Final

from .base_event import OpenhabEvent
from ..definitions import ThingStatusEnum, ThingStatusDetailEnum
from ..definitions.things import THING_STATUS_DEFAULT, THING_STATUS_DETAIL_DEFAULT


class ThingStatusInfoEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar ThingStatusEnum status:
    :ivar ThingStatusDetailEnum detail:
    :ivar str description:
    """
    name: str
    status: ThingStatusEnum
    detail: ThingStatusDetailEnum
    description: str

    def __init__(self, name: str = '', status: ThingStatusEnum = THING_STATUS_DEFAULT,
                 detail: ThingStatusDetailEnum = THING_STATUS_DETAIL_DEFAULT, description: str = ''):
        super().__init__()

        self.name: Final = name
        self.status: Final = status
        self.detail: Final = detail
        self.description: Final = description

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # openhab/things/chromecast:chromecast:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/status
        return cls(name=topic[15:-7], status=ThingStatusEnum(payload['status']),
                   detail=ThingStatusDetailEnum(payload['statusDetail']), description=payload.get('description', ''))

    def __repr__(self):
        description = f', description: "{self.description:s}"' if self.description else ''
        return f'<{self.__class__.__name__} name: {self.name:s}, ' \
               f'status: {self.status:s}, detail: {self.detail:s}{description:s}>'

    # def __init__(self, name: str = '', status: ThingStatusEnum = THING_STATUS_DEFAULT,
    #              detail: ThingStatusDetailEnum = THING_STATUS_DETAIL_DEFAULT, description: str = ''):


class ThingStatusInfoChangedEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar ThingStatusEnum status:
    :ivar ThingStatusDetailEnum detail:
    :ivar str description:
    :ivar ThingStatusEnum old_status:
    :ivar ThingStatusDetailEnum old_detail:
    :ivar str old_description:
    """
    name: str
    status: ThingStatusEnum
    detail: ThingStatusDetailEnum
    description: str
    old_status: ThingStatusEnum
    old_detail: ThingStatusDetailEnum
    old_description: str

    def __init__(self, name: str = '',
                 status: ThingStatusEnum = THING_STATUS_DEFAULT,
                 detail: ThingStatusDetailEnum = THING_STATUS_DETAIL_DEFAULT,
                 description: str = '',
                 old_status: ThingStatusEnum = THING_STATUS_DEFAULT,
                 old_detail: ThingStatusDetailEnum = THING_STATUS_DETAIL_DEFAULT,
                 old_description: str = ''):
        super().__init__()

        self.name: Final = name
        self.status: Final = status
        self.detail: Final = detail
        self.description: Final = description

        self.old_status: Final = old_status
        self.old_detail: Final = old_detail
        self.old_description: Final = old_description

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # openhab/things/chromecast:chromecast:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa/statuschanged
        name = topic[15:-14]
        new, old = payload
        return cls(
            name=name,
            status=ThingStatusEnum(new['status']), detail=ThingStatusDetailEnum(new['statusDetail']),
            description=new.get('description', ''),
            old_status=ThingStatusEnum(old['status']), old_detail=ThingStatusDetailEnum(old['statusDetail']),
            old_description=old.get('description', '')
        )

    def __repr__(self):
        description = f', description: "{self.description:s}"' if self.description else ''
        old_description = f', old_description: "{self.old_description:s}"' if self.old_description else ''
        return f'<{self.__class__.__name__} name: {self.name}, ' \
               f'status: {self.status}, detail: {self.detail}{description:s}, ' \
               f'old_status: {self.old_status}, old_detail: {self.old_detail}{old_description:s}>'


class ThingConfigStatusInfoEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar Dict[str, str] config_messages:
    """
    name: str
    config_messages: Dict[str, str]

    def __init__(self, name: str = '', config_messages: Dict[str, str] = None):
        super().__init__()

        self.name: str = name
        self.config_messages: Dict[str, str] = config_messages if config_messages is not None else {}

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # openhab/things/zwave:device:gehirn:node29/config/status
        name = topic[15:-14]
        msgs = payload['configStatusMessages']
        return cls(
            name=name, config_messages={param_name: msg_type for d in msgs for param_name, msg_type in d.items()}
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, config_messages: {self.config_messages}>'


class ThingFirmwareStatusInfoEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str status:
    """
    name: str
    status: str

    def __init__(self, name: str = '', status: str = ''):
        super().__init__()
        self.name: str = name
        self.status: str = status

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'openhab/things/zwave:device:controller:my_node/firmware/status'
        return cls(name=topic[15:-16], status=payload['firmwareStatus'])

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name} status: {self.status}>'


class ThingRegistryBaseEvent(OpenhabEvent):
    """
    :ivar str name:
    :ivar str type:
    :ivar str label:
    :ivar str location:
    :ivar List[Dict[str, Any]] channels:
    :ivar Dict[str, Any] configuration:
    :ivar Dict[str, str] properties:
    """
    name: str
    type: str
    label: str
    location: str
    channels: List[Dict[str, Any]]
    configuration: Dict[str, Any]
    properties: Dict[str, str]

    def __init__(self, name: str, thing_type: str, label: str, location: str,
                 channels: List[Dict[str, Any]], configuration: Dict[str, Any],
                 properties: Dict[str, str]):
        super().__init__()

        # use name instead of uuid
        self.name: Final = name
        self.type: Final = thing_type

        # optional entries
        self.label: Final = label
        self.location: Final = location
        self.channels: Final = channels
        self.configuration: Final = configuration
        self.properties: Final = properties

    @classmethod
    def from_dict(cls, topic: str, payload: dict):
        # 'openhab/things/astro:sun:0a94363608/added'
        return cls(
            name=payload['UID'], thing_type=payload['thingTypeUID'], label=payload['label'],
            location=payload.get('location', ''),
            channels=payload.get('channels', []),
            configuration=payload.get('configuration', {}),
            properties=payload.get('properties', {}),
        )

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}>'


class ThingAddedEvent(ThingRegistryBaseEvent):
    pass


class ThingRemovedEvent(ThingRegistryBaseEvent):
    pass


class ThingUpdatedEvent(ThingRegistryBaseEvent):
    @classmethod
    def from_dict(cls, topic: str, payload: List[Dict[str, Any]]):

        payload = payload[0]
        return cls(
            name=payload['UID'], thing_type=payload['thingTypeUID'], label=payload['label'],
            location=payload.get('location', ''),
            channels=payload.get('channels'),
            configuration=payload.get('configuration'),
            properties=payload.get('properties'),
        )



import ast
import datetime
from inspect import ismodule, isclass
from pathlib import Path
from typing import List, Tuple, Callable, Any, Set

from immutables import Map
from stack_data import Variable

from HABApp.core.const.json import load_json, dump_json
from easyconfig.config_objs import ConfigObj
from .const import SEPARATOR_VARIABLES, PRE_INDENT

# don't show these types in the traceback
SKIPPED_TYPES = (
    bool, bytearray, bytes, complex, dict, float, frozenset, int, list, memoryview, set, str, tuple, type(None),
    datetime.date, datetime.datetime, datetime.time, datetime.timedelta,
    Map, Path,
)


def is_type_hint_or_type(value: Any) -> bool:
    if isinstance(value, tuple):
        return all(is_type_hint_or_type(obj) for obj in value)

    if value in SKIPPED_TYPES:
        return True

    # check if it's something from the typing module
    # we can't do that with isinstance, so we try this
    try:
        if value.__module__ == 'typing' or value.__class__.__module__ == 'typing':
            return True
    except AttributeError:
        pass

    return False


def _filter_expressions(name: str, value: Any) -> bool:
    # a is None = True
    if name.endswith(' is None'):
        return True

    return False


SKIP_VARIABLE: Tuple[Callable[[str, Any], bool], ...] = (
    # module imports
    lambda name, value: ismodule(value),

    # type hints and type tuples
    lambda name, value: is_type_hint_or_type(value),

    # functions
    lambda name, value: value is dump_json or value is load_json,

    # config value objs
    lambda name, value: isinstance(value, ConfigObj),

    # Expressions
    _filter_expressions
)

ORDER_VARIABLE: Tuple[Callable[[Variable], bool], ...] = (
    lambda x: isclass(x.value),
)


def skip_variable(var: Variable) -> bool:
    name = var.name
    value = var.value
    for func in SKIP_VARIABLE:
        if func(name, value):
            return True
    return False


def format_frame_variables(tb: List[str], stack_variables: List[Variable]):
    if not stack_variables:
        return None

    # remove variables that shall not be printed
    used_vars: List[Variable] = [v for v in stack_variables if not skip_variable(v)]

    # attributes
    dotted_names: Set[str] = {n.name.split('.')[0] for n in used_vars if '.' in n.name}

    # Sort output
    used_vars = sorted(used_vars, key=lambda x: (
        isinstance(x.nodes[0], ast.Compare),                                        # Compare objects last
        not any(map(
            lambda y: x.name == y or x.name.startswith(y + '.'), dotted_names)),    # Classes with attributes
        x.name.lower()                                                              # Name lowercase
    ))

    variables = {}

    # variables by order
    for add_var in ORDER_VARIABLE:
        for var in used_vars:
            name = var.name
            if name in variables:
                continue
            if add_var(var):
                variables[name] = var.value

    # rest of the variables
    for var in used_vars:
        if var.name not in variables:
            variables[var.name] = var.value

    if not variables:
        return None

    # Add variables to traceback
    tb.append(SEPARATOR_VARIABLES)

    for name, value in variables.items():
        tb.append(f'{" " * (PRE_INDENT + 1)}{name} = {repr(value)}')

    tb.append(SEPARATOR_VARIABLES)



import logging
import typing

import HABApp
from . import ValueMode


class SwitchItemValueMode(ValueMode):
    """SwitchItemMode, same as ValueMode but enabled/disabled of the mode is controlled by a OpenHAB
    :class:`~HABApp.openhab.items.SwitchItem`

    :ivar datetime.datetime last_update: Timestamp of the last update/enable of this value
    :ivar typing.Optional[datetime.timedelta] auto_disable_after: Automatically disable this mode after
                                                                    a given timedelta on the next recalculation
    :vartype auto_disable_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], bool]]
    :ivar    auto_disable_func: Function which can be used to disable this mode. Any function that accepts two
                                Arguments can be used. First arg is value with lower priority,
                                second argument is own value. Return ``True`` to disable this mode.
    :vartype calc_value_func: typing.Optional[typing.Callable[[typing.Any, typing.Any], typing.Any]]
    :ivar    calc_value_func: Function to calculate the new value (e.g. ``min`` or ``max``). Any function that accepts
                              two Arguments can be used. First arg is value with lower priority,
                              second argument is own value.
    """

    def __init__(self, name: str,
                 # these are the parameters special to SwitchItemValueMode
                 switch_item: 'HABApp.openhab.items.SwitchItem', invert_switch: bool = False,
                 # default kw-args from the base class
                 initial_value=None,
                 logger: typing.Optional[logging.Logger] = None,
                 auto_disable_after=None, auto_disable_func=None,
                 calc_value_func=None):
        """

        :param name: Name of the mode
        :param switch_item: :class:`~HABApp.openhab.items.SwitchItem` that will enable/disable the mode
        :param invert_switch: invert switch state (e.g. `OFF` -> enabled, default is ``False``)
        :param initial_value: initial value of the mode
        :param logger: logger to log events
        :param auto_disable_after: see variables
        :param auto_disable_func: see variables
        :param calc_value_func: see variables
        """

        assert invert_switch is True or invert_switch is False
        assert isinstance(switch_item, HABApp.openhab.items.SwitchItem), type(switch_item)
        self.__invert_switch: bool = invert_switch

        super().__init__(name=name,
                         initial_value=initial_value,
                         enabled=switch_item.value == ('ON' if not self.__invert_switch else 'OFF'),
                         enable_on_value=False,  # enable_on_value must be pinned False
                         logger=logger,
                         auto_disable_after=auto_disable_after, auto_disable_func=auto_disable_func,
                         calc_value_func=calc_value_func)

        # setup listener as the last thing
        switch_item.listen_event(self.__switch_changed, HABApp.core.events.ValueChangeEventFilter())
        return

    # this is the original enabled method
    __set_enable = ValueMode.set_enabled

    # prevent direct calling
    def set_enabled(self, value: bool, only_on_change: bool = False):
        """"""  # Empty docstring so this function doesn't show up in Sphinx
        raise PermissionError('Enabled is controlled through the switch item!')

    def __switch_changed(self, event):
        self.__set_enable(event.value == ('ON' if not self.__invert_switch else 'OFF'))



from __future__ import annotations

from typing import Any

import aiohttp
from aiohttp.client import ClientResponse, _RequestContextManager
from aiohttp.hdrs import METH_GET, METH_POST, METH_PUT, METH_DELETE

from HABApp.config import CONFIG
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.connections._definitions import CONNECTION_HANDLER_NAME
from HABApp.core.connections.base_connection import AlreadyHandledException
from HABApp.core.const.json import dump_json
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.errors import OpenhabDisconnectedError, OpenhabCredentialsInvalidError


# noinspection PyProtectedMember
class ConnectionHandler(BaseConnectionPlugin[OpenhabConnection]):
    request: aiohttp.ClientSession._request

    def __init__(self):
        super().__init__(name=CONNECTION_HANDLER_NAME)
        self.options: dict[str, Any] = {}
        self.read_only: bool = False
        self.online = False
        self.session: aiohttp.ClientSession | None = None

    def update_cfg_general(self):
        self.read_only = CONFIG.openhab.general.listen_only

    async def on_setup(self, connection: OpenhabConnection):
        log = self.plugin_connection.log
        config = CONFIG.openhab.connection
        url: str = config.url
        user: str = config.user
        password: str = config.password

        # do not run without an url
        if not url:
            log.info('Connection disabled (url missing)!')
            connection.status_from_setup_to_disabled()
            return None

        # do not run without user/pw - since OH3 mandatory
        is_token = user.startswith('oh.') or password.startswith('oh.')
        if not is_token and (not user or not password):
            log.info('Connection disabled (user/password missing)!')
            connection.status_from_setup_to_disabled()
            return None

        if not config.verify_ssl:
            self.options['ssl'] = False
            log.info('Verify ssl set to False!')
        else:
            self.options.pop('ssl', None)

        self.update_cfg_general()

        # remove existing session
        if (s := self.session) is not None:
            self.session = None
            await s.close()

        self.session = aiohttp.ClientSession(
            base_url=url,
            timeout=aiohttp.ClientTimeout(total=None),
            json_serialize=dump_json,
            auth=aiohttp.BasicAuth(user, password),
            read_bufsize=int(config.buffer)
        )
        self.request = self.session._request

    async def on_connected(self):
        self.online = True

    async def on_disconnected(self, connection: OpenhabConnection):
        self.online = False
        connection.context = None

    async def on_shutdown(self):
        if self.session is None:
            return None

        await self.session.close()
        self.session = None

    async def get(self, url: str, log_404=True, **kwargs: Any) -> ClientResponse:
        mgr = _RequestContextManager(self.request(METH_GET, url, **self.options, **kwargs))
        return await self.check_response(mgr, log_404=log_404)

    async def post(self, url: str, log_404=True, json=None, data=None, **kwargs: Any) -> ClientResponse | None:
        if self.read_only or not self.online:
            return None

        mgr = _RequestContextManager(
            self.request(METH_POST, url, data=data, json=json, **kwargs, **self.options, **kwargs)
        )
        if data is None:
            data = json
        return await self.check_response(mgr, log_404=log_404, sent_data=data)

    async def put(self, url: str, log_404=True, json=None, data=None, **kwargs: Any) -> ClientResponse | None:
        if self.read_only or not self.online:
            return None

        mgr = _RequestContextManager(
            self.request(METH_PUT, url, data=data, json=json, **kwargs, **self.options, **kwargs)
        )
        if data is None:
            data = json
        return await self.check_response(mgr, log_404=log_404, sent_data=data)

    async def delete(self, url: str, log_404=True, json=None, data=None, **kwargs: Any) -> ClientResponse | None:
        if self.read_only or not self.online:
            return None

        mgr = _RequestContextManager(
            self.request(METH_DELETE, url, data=data, json=json, **kwargs, **self.options, **kwargs)
        )
        if data is None:
            data = json
        return await self.check_response(mgr, log_404=log_404, sent_data=data)

    async def check_response(self, future: aiohttp.client._RequestContextManager, sent_data=None,
                             log_404=True) -> ClientResponse:
        try:
            resp = await future
        except Exception as e:
            self.plugin_connection.process_exception(e, None)
            raise OpenhabDisconnectedError() from None

        if (status := resp.status) < 300:
            return resp

        if status == 404 and not log_404:
            return resp

        # Log Error Message
        log = self.plugin_connection.log
        sent = '' if sent_data is None else f' {sent_data}'
        log.warning(f'Status {status} for {resp.request_info.method} {resp.request_info.url}{sent}')
        for line in str(resp).splitlines():
            log.debug(line)

        if resp.status == 401:
            raise OpenhabCredentialsInvalidError()

        return resp

    async def on_connecting(self, connection: OpenhabConnection):
        from HABApp.openhab.connection.handler.func_async import async_get_root

        log = self.plugin_connection.log
        log.debug('Trying to connect to OpenHAB ...')

        try:
            if (root := await async_get_root()) is None:
                connection.set_error()
                log.info('... offline!')
                return None

            info = root.runtime_info
            log.info(f'Connected {"read only " if self.read_only else ""}to OpenHAB '
                     f'version {info.version:s} ({info.build_string:s})')

            vers = tuple(map(int, info.version.split('.')[:3]))
            if vers < (3, 3):
                log.warning('HABApp requires at least openHAB version 3.3!')

            connection.context = OpenhabContext(
                version=vers, is_oh3=vers < (4, 0),
                waited_for_openhab=False,
                created_items={}, created_things={},
                session=self.session, session_options=self.options
            )

        # during startup we get OpenhabCredentialsInvalidError even though credentials are correct
        except (OpenhabDisconnectedError, OpenhabCredentialsInvalidError):
            connection.set_error()
            raise AlreadyHandledException()


HANDLER = ConnectionHandler()
get = HANDLER.get
post = HANDLER.post
put = HANDLER.put
delete = HANDLER.delete



import logging
from typing import Dict, Set, Tuple, TYPE_CHECKING, Union

from immutables import Map

import HABApp

from HABApp.core.internals import uses_item_registry
from HABApp.core.logger import log_warning
from HABApp.openhab.definitions.things import THING_STATUS_DEFAULT, THING_STATUS_DETAIL_DEFAULT, ThingStatusEnum, \
    ThingStatusDetailEnum

if TYPE_CHECKING:
    import HABApp.openhab.definitions.rest

log = logging.getLogger('HABApp.openhab.items')

Items = uses_item_registry()


def add_to_registry(item: 'HABApp.openhab.items.OpenhabItem', set_value=False):
    name = item.name
    for grp in item.groups:
        MEMBERS.setdefault(grp, set()).add(name)

    if not Items.item_exists(name):
        return Items.add_item(item)

    existing = Items.get_item(name)
    if isinstance(existing, item.__class__):
        # If we load directly through the API and not through an event we have to set the value
        if set_value:
            existing.set_value(item.value)

        # remove old groups
        for grp in set(existing.groups) - set(item.groups):
            MEMBERS.get(grp, set()).discard(name)

        # same type - it was only an item update (e.g. label)!
        existing.label    = item.label
        existing.tags     = item.tags
        existing.groups   = item.groups
        existing.metadata = item.metadata
        return None

    log_warning(log, f'Item type changed from {existing.__class__} to {item.__class__}')

    # Replace existing item with the updated definition
    Items.pop_item(name)
    Items.add_item(item)


def remove_from_registry(name: str):
    if not Items.item_exists(name):
        return None

    item = Items.get_item(name)  # type: HABApp.openhab.items.OpenhabItem
    for grp in item.groups:
        MEMBERS.get(grp, set()).discard(name)

    if isinstance(item, HABApp.openhab.items.GroupItem):
        MEMBERS.pop(name, None)

    Items.pop_item(name)
    return None


MEMBERS: Dict[str, Set[str]] = {}


def fresh_item_sync():
    MEMBERS.clear()


def get_members(group_name: str) -> Tuple['HABApp.openhab.items.OpenhabItem', ...]:
    ret = []
    for name in MEMBERS.get(group_name, []):
        item = Items.get_item(name)  # type: HABApp.openhab.items.OpenhabItem
        ret.append(item)
    return tuple(sorted(ret, key=lambda x: x.name))


# ----------------------------------------------------------------------------------------------------------------------
# Thing handling
# ----------------------------------------------------------------------------------------------------------------------
def add_thing_to_registry(data: Union['HABApp.openhab.definitions.rest.ThingResp',
                                      'HABApp.openhab.events.thing_events.ThingAddedEvent']):

    if isinstance(data, HABApp.openhab.events.thing_events.ThingAddedEvent):
        name = data.name
        status: ThingStatusEnum = THING_STATUS_DEFAULT
        status_detail: ThingStatusDetailEnum = THING_STATUS_DETAIL_DEFAULT
        status_description: str = ''
    elif isinstance(data, HABApp.openhab.definitions.rest.ThingResp):
        name = data.uid
        status = data.status.status
        status_detail = data.status.detail
        status_description = data.status.description if data.status.description else ''
    else:
        raise ValueError()

    if Items.item_exists(name):
        existing = Items.get_item(name)
        if isinstance(existing, HABApp.openhab.items.Thing):
            new_thing = existing
        else:
            # Replace existing item with the correct type
            new_thing = HABApp.openhab.items.Thing(name=name)
            log_warning(log, f'Item type changed from {existing.__class__} to {new_thing.__class__}')
            Items.pop_item(name)
    else:
        new_thing = HABApp.openhab.items.Thing(name=name)

    new_thing.status = status
    new_thing.status_detail = status_detail
    new_thing.status_description = status_description
    new_thing.label = data.label
    new_thing.location = data.location
    new_thing.configuration = Map(data.configuration)
    new_thing.properties = Map(data.properties)
    return Items.add_item(new_thing)


def remove_thing_from_registry(name: str):
    if not Items.item_exists(name):
        return None
    return Items.pop_item(name)



import logging
from typing import Dict, Any, Final, TypeVar, Tuple, Generic


T = TypeVar('T')

log = logging.getLogger('HABApp.openhab.transform')


class TransformationFactoryBase(Generic[T]):
    def __init__(self, registry: 'TransformationRegistryBase'):
        self._registry: Final = registry

    def __repr__(self):
        return f'<{self._registry.name.title()}{self.__class__.__name__}>'

    def __getitem__(self, key: str) -> T:
        return self._registry.get(key)


def sort_order(uid: str):
    # created through file
    if '.' in uid:
        name, ext = uid.rsplit('.', 1)
        return 0, ext, name

    # UI created
    return 1, uid.split(':')


class TransformationRegistryBase:
    objs: Dict[str, Any]

    def __init__(self, name: str):
        self.name: Final = name

    def __repr__(self):
        return f'<{self.__class__.__name__} {" ".join(self.available())}'

    def available(self) -> Tuple[str, ...]:
        return tuple(sorted(self.objs.keys(), key=sort_order))

    def get(self, name: str):
        raise NotImplementedError()

    def set(self, name: str, configuration: dict):
        raise NotImplementedError()

    def clear(self):
        self.objs.clear()



import re
from typing import List

from stack_data import FrameInfo, LINE_GAP

from .const import SEPARATOR_NEW_FRAME, PRE_INDENT
from .format_frame_vars import format_frame_variables

SUPPRESSED_HABAPP_PATHS = (
    # This exception formatter
    re.compile(r'[/\\]HABApp[/\\]core[/\\]lib[/\\]exceptions[/\\]'),
    # Wrapper which usually generates this traceback
    re.compile(r'[/\\]HABApp[/\\]core[/\\]wrapper.py'),

    # Rule file loader
    re.compile(r'[/\\]HABApp[/\\]rule_manager[/\\]'),

    # Worker functions
    re.compile(r'[/\\]HABApp[/\\]core[/\\]internals[/\\]wrapped_function[/\\]'),

    # Item registry
    re.compile(r'[/\\]HABApp[/\\]core[/\\]internals[/\\]item_registry[/\\]'),

    # Connection wrappers
    re.compile(r'[/\\]HABApp[/\\]core[/\\]connections[/\\]'),
)

SUPPRESSED_PATHS = (
    # Libraries of base installation
    re.compile(r'[/\\](?:python\d\.\d+|python\d{2,3})[/\\](?:lib[/\\]|site-packages[/\\]|\w+\.py.*$)', re.IGNORECASE),
    # Libraries in venv
    re.compile(r'[/\\]lib[/\\]site-packages[/\\]', re.IGNORECASE),
)


def is_suppressed_habapp_file(name: str) -> bool:
    for r in SUPPRESSED_HABAPP_PATHS:
        if r.search(name):
            return True
    return False


def is_lib_file(name: str) -> bool:
    for r in SUPPRESSED_PATHS:
        if r.search(name):
            if '/HABApp/' in name or '\\HABApp\\' in name:
                continue
            return True
    return False


def format_frame_info(tb: List[str], frame_info: FrameInfo, is_last=False) -> bool:
    filename = frame_info.filename

    # always skip system and python libraries
    if is_lib_file(filename):
        return False

    if not is_last and is_suppressed_habapp_file(filename):
        return False

    # calc max line nr for indentation
    max_line = frame_info.lineno + frame_info.options.after
    # it's possible that his list is empty (e.g. if we execode in sphinx-exec-code)
    if frame_info.lines:
        max_line = frame_info.lines[-1].lineno

    indent = len(str(max_line)) + 1

    tb.append(f'File "{filename}", line {frame_info.lineno} in {frame_info.code.co_name}')
    tb.append(SEPARATOR_NEW_FRAME)
    for line in frame_info.lines:
        if line is LINE_GAP:
            tb.append(f"{' ' * (PRE_INDENT + indent - 1)}(...)")
        else:
            tb.append(f"{'-->' if line.is_current else '':{PRE_INDENT}s}{line.lineno:{indent}d} | {line.render()}")

    format_frame_variables(tb, frame_info.variables)
    tb.append('')
    return True



from typing import Any, Union, Final


class ComplexEventValue:
    def __init__(self, value):
        self.value: Any = value


class ValueUpdateEvent:
    """
    :ivar str name:
    :ivar Any value:
    """

    name: str
    value: Any

    def __init__(self, name: str, value: Any):
        self.name: Final = name
        self.value: Final = value

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}>'


class ValueChangeEvent:
    """
    :ivar str name:
    :ivar Any value:
    :ivar Any old_value:
    """

    name: str
    value: Any
    old_value: Any

    def __init__(self, name: str, value: Any, old_value: Any):
        self.name: Final = name
        self.value: Final = value
        self.old_value: Final = old_value

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, value: {self.value}, old_value: {self.old_value}>'


class ItemNoChangeEvent:
    """
    :ivar str               name:
    :ivar Union[int, float] seconds:
    """

    name: str
    seconds: Union[int, float]

    def __init__(self, name: str, seconds: Union[int, float]):
        self.name: Final = name
        self.seconds: Final = seconds

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, seconds: {self.seconds}>'


class ItemNoUpdateEvent:
    """
    :ivar str               name:
    :ivar Union[int, float] seconds:
    """
    name: str
    seconds: Union[int, float]

    def __init__(self, name: str, seconds: Union[int, float]):
        self.name: Final = name
        self.seconds: Final = seconds

    def __repr__(self):
        return f'<{self.__class__.__name__} name: {self.name}, seconds: {self.seconds}>'



import io
import logging
from cProfile import Profile
from concurrent.futures import ThreadPoolExecutor
from pstats import SortKey
from pstats import Stats
from threading import Lock
from time import monotonic
from typing import Callable, Any, Set, Final, Dict, Tuple
from typing import Optional

from HABApp.core.const import loop
from HABApp.core.internals import Context, ContextProvidingObj
from .base import WrappedFunctionBase, default_logger

POOL: Optional[ThreadPoolExecutor] = None
POOL_THREADS: int = 0
POOL_INFO: Set['PoolFunc'] = set()
POOL_LOCK = Lock()


def create_thread_pool(count: int):
    global POOL, POOL_THREADS
    assert isinstance(count, int) and count > 0

    default_logger.debug(f'Starting thread pool with {count:d} threads!')

    stop_thread_pool()
    POOL_THREADS = count
    POOL = ThreadPoolExecutor(count, 'HabAppWorker')


def stop_thread_pool():
    global POOL, POOL_THREADS
    if POOL is not None:
        POOL.shutdown()
        POOL = None
        POOL_THREADS = 0
        default_logger.debug('Thread pool stopped!')


async def run_in_thread_pool(func: Callable):
    return await loop.run_in_executor(
        POOL, func
    )


HINT_FUNC_SYNC = Callable[..., Any]


class PoolFunc(ContextProvidingObj):
    def __init__(self, parent: 'WrappedThreadFunction', func_obj: HINT_FUNC_SYNC, func_args: Tuple[Any, ...],
                 func_kwargs: Dict[str, Any], context: Optional[Context] = None, **kwargs):
        super().__init__(context=context, **kwargs)
        self.parent: Final = parent
        self.func_obj: Final = func_obj
        self.func_args: Final = func_args
        self.func_kwargs: Final = func_kwargs

        # timing checks
        self.submitted = monotonic()
        self.dur_start = 0.0
        self.dur_run = 0.0

        # thread info
        self.usage_high = 0

    def __repr__(self):
        return f'<{self.__class__.__name__} high: {self.usage_high:d}/{POOL_THREADS:d}>'

    def run(self):
        try:
            ts_start = monotonic()
            self.dur_start = ts_start - self.submitted

            with POOL_LOCK:
                POOL_INFO.add(self)
                count = len(POOL_INFO)
                for info in POOL_INFO:
                    info.usage_high = max(count, info.usage_high)

            parent = self.parent

            # notify if we don't process quickly
            if self.dur_start > 0.05:
                parent.log.warning(f'Starting of {parent.name} took too long: {self.dur_start:.2f}s. '
                                   f'Maybe there are not enough threads?')

            # start profiler
            pr = Profile()
            pr.enable()

            # Execute the function
            self.func_obj(*self.func_args, **self.func_kwargs)

            # disable profiler
            pr.disable()

            # log warning if execution takes too long
            self.dur_run = monotonic() - ts_start

            if parent.warn_too_long and self.dur_run > 0.8 and self.usage_high >= POOL_THREADS * 0.6:
                parent.log.warning(f'{self.usage_high}/{POOL_THREADS} threads have been in use and '
                                   f'execution of {parent.name} took too long: {self.dur_run:.2f}s')

                s = io.StringIO()
                ps = Stats(pr, stream=s).sort_stats(SortKey.CUMULATIVE)
                ps.print_stats(0.1)  # limit to output to 10% of the lines

                for line in s.getvalue().splitlines()[4:]:  # skip the amount of calls and "Ordered by:"
                    if line:
                        parent.log.warning(line)

        except Exception as e:
            self.parent.process_exception(e, *self.func_args, **self.func_kwargs)
            return None
        finally:
            with POOL_LOCK:
                POOL_INFO.discard(self)


class WrappedThreadFunction(WrappedFunctionBase):

    def __init__(self, func: HINT_FUNC_SYNC,
                 warn_too_long=True,
                 name: Optional[str] = None,
                 logger: Optional[logging.Logger] = None,
                 context: Optional[Context] = None):

        super().__init__(name=name, func=func, logger=logger, context=context)
        assert callable(func)

        self.func = func
        self.warn_too_long: bool = warn_too_long

    def run(self, *args, **kwargs):
        # we need to copy the context, so it's available when the function is run
        pool_func = PoolFunc(self, self.func, args, kwargs, context=self._habapp_ctx)
        POOL.submit(pool_func.run)



from __future__ import annotations

import logging
from typing import Final

from aiohttp_sse_client import client as sse_client

import HABApp
import HABApp.core
import HABApp.openhab.events
from HABApp.core.asyncio import AsyncContext
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.const.json import load_json
from HABApp.core.const.log import TOPIC_EVENTS
from HABApp.core.internals import uses_item_registry
from HABApp.core.lib import SingleTask
from HABApp.openhab.connection.connection import OpenhabConnection
from HABApp.openhab.process_events import on_sse_event

Items = uses_item_registry()


class SseEventListenerPlugin(BaseConnectionPlugin[OpenhabConnection]):

    def __init__(self, name: str | None = None):
        super().__init__(name)
        self.task: Final = SingleTask(self.sse_task, name='SSE Task')

    async def on_connected(self):
        self.task.start()

    async def on_disconnected(self):
        await self.task.cancel_wait()

    async def sse_task(self):
        try:
            with AsyncContext('SSE'):
                # cache so we don't have to look up every event
                _load_json = load_json
                _see_handler = on_sse_event
                context = self.plugin_connection.context
                oh3 = context.is_oh3

                log_events = logging.getLogger(f'{TOPIC_EVENTS}.openhab')
                DEBUG = logging.DEBUG

                async with sse_client.EventSource(
                        url=f'/rest/events?topics={HABApp.CONFIG.openhab.connection.topic_filter}',
                        session=context.session, **context.session_options) as event_source:
                    async for event in event_source:

                        e_str = event.data

                        try:
                            e_json = _load_json(e_str)
                        except (ValueError, TypeError):
                            log_events.warning(f'Invalid json: {e_str}')
                            continue

                        # Alive event from openhab to detect dropped connections
                        # -> Can be ignored on the HABApp side
                        e_type = e_json.get('type')
                        if e_type == 'ALIVE':
                            continue

                        # Log raw sse event
                        if log_events.isEnabledFor(logging.DEBUG):
                            log_events._log(DEBUG, e_str, [])

                        # With OH4 we have the ItemStateUpdatedEvent, so we can ignore the ItemStateEvent
                        if not oh3 and e_type == 'ItemStateEvent':
                            continue

                        # process
                        _see_handler(e_json, oh3)
        except Exception as e:
            self.plugin_connection.process_exception(e, self.sse_task)



from typing import Optional, List

from msgspec import Struct, field


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.persistence/src/main/java/org/openhab/core/persistence/dto/PersistenceServiceDTO.java
class PersistenceServiceResp(Struct):
    id: str
    label: Optional[str] = None
    type: Optional[str] = None


class DataPoint(Struct):
    time: int
    state: str


# ItemHistoryDTO
# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.persistence/src/main/java/org/openhab/core/persistence/dto/ItemHistoryDTO.java

class ItemHistoryResp(Struct):
    name: str
    total_records: Optional[str] = field(default=None, name='totalrecords')
    data_points: Optional[str] = field(default=None, name='datapoints')
    data: List[DataPoint] = []



from typing import Union

from pydantic import AnyHttpUrl, ByteSize, Field, field_validator, TypeAdapter

from easyconfig.models import BaseModel


class Ping(BaseModel):
    enabled: bool = Field(True, description='If enabled the configured item will show how long it takes to send '
                                            'an update from HABApp and get the updated value back from openHAB '
                                            'in milliseconds')
    item: str = Field('HABApp_Ping', description='Name of the Numberitem')
    interval: Union[int, float] = Field(10, description='Seconds between two pings', ge=0.1)


class General(BaseModel):
    listen_only: bool = Field(
        False, description='If True HABApp does not change anything on the openHAB instance.'
    )
    wait_for_openhab: bool = Field(
        True,
        description='If True HABApp will wait for a successful openHAB connection before loading any rules on startup'
    )

    # Advanced settings
    min_start_level: int = Field(
        70, ge=0, le=100, in_file=False,
        description='Minimum openHAB start level to load items and listen to events',
    )


class Connection(BaseModel):
    url: str = Field(
        'http://localhost:8080', description='Connect to this url. Empty string ("") disables the connection.')
    user: str = ''
    password: str = ''
    verify_ssl: bool = Field(True, description='Check certificates when using https')

    buffer: ByteSize = Field(
        '128kib', in_file=False, description=
        'Buffer for reading lines in the SSE event handler. This is the buffer '
        'that gets allocated for every(!) request and SSE message that the client processes. '
        'Increase only if you get error messages or disconnects e.g. if you use large images.'
    )

    topic_filter: str = Field(
        'openhab/items/*,'      # Item updates
        'openhab/channels/*,'   # Channel update
        # Thing events - don't listen to updated events
        # todo: check if this might be a good filter: 'openhab/things/*',
        'openhab/things/*',
        alias='topic filter', in_file=False,
        description='Topic filter for subscribing to openHAB. This filter is processed by openHAB and only events '
                    'matching this filter will be sent to HABApp.'
    )

    @field_validator('url')
    def validate_url(cls, value: str):
        TypeAdapter(AnyHttpUrl).validate_python(value)
        return value

    @field_validator('buffer')
    def validate_see_buffer(cls, value: ByteSize):
        valid_values = (
            '64kib', '128kib', '256kib', '512kib',
            '1Mib', '2Mib', '4Mib', '8Mib', '16Mib', '32Mib', '64Mib', '128Mib'
        )

        for _v in valid_values:
            # noinspection PyProtectedMember
            if value == ByteSize._validate(_v, None):
                return value

        raise ValueError(f'Value must be one of {", ".join(valid_values)}')


class OpenhabConfig(BaseModel):
    connection: Connection = Field(default_factory=Connection)
    general: General = Field(default_factory=General)
    ping: Ping = Field(default_factory=Ping)



import logging.config

from pydantic import Field

from easyconfig import BaseModel

log = logging.getLogger('HABApp.Config')


class LocationConfig(BaseModel):
    """location where the instance is running. Is used to calculate Sunrise/Sunset."""

    latitude: float = Field(default=0.0)
    longitude: float = Field(default=0.0)
    elevation: float = Field(default=0.0)



from HABApp.__check_dependency_packages__ import check_dependency_packages

# -----------------------------------------------------------------------------
# if installed we use uvloop because it seems to be much faster (untested)
# -----------------------------------------------------------------------------
try:
    import uvloop   # type: ignore

    uvloop.install()
    print('Using uvloop')
except ModuleNotFoundError:
    pass


# -----------------------------------------------------------------------------
# Check that all dependencies are installed
# We do this here, so we can print a nice error message. Otherwise the corresponding
# module import will fail somewhere in the middle of the startup process
# -----------------------------------------------------------------------------
check_dependency_packages()



from __future__ import annotations

import logging
from typing import Optional

from aiomqtt import Client, MqttError

import HABApp
from HABApp.core.asyncio import AsyncContext
from HABApp.core.connections import BaseConnection, Connections, ConnectionStateToEventBusPlugin, AutoReconnectPlugin
from HABApp.core.connections.base_plugin import BaseConnectionPluginConnectedTask
from HABApp.core.const.const import PYTHON_310

log = logging.getLogger('HABApp.mqtt.connection')

if PYTHON_310:
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

CONTEXT_TYPE: TypeAlias = Optional[Client]


def setup():
    config = HABApp.config.CONFIG.mqtt

    from HABApp.mqtt.connection.handler import CONNECTION_HANDLER
    from HABApp.mqtt.connection.subscribe import SUBSCRIPTION_HANDLER
    from HABApp.mqtt.connection.publish import PUBLISH_HANDLER

    connection = Connections.add(MqttConnection())

    connection.register_plugin(CONNECTION_HANDLER, 0)
    connection.register_plugin(SUBSCRIPTION_HANDLER, 10)
    connection.register_plugin(PUBLISH_HANDLER, 20)

    connection.register_plugin(ConnectionStateToEventBusPlugin())
    connection.register_plugin(AutoReconnectPlugin())

    # config changes
    config.subscribe.subscribe_for_changes(SUBSCRIPTION_HANDLER.subscription_cfg_changed)
    config.connection.subscribe_for_changes(connection.status_configuration_changed)


class MqttConnection(BaseConnection):
    def __init__(self):
        super().__init__('mqtt')
        self.context: CONTEXT_TYPE = None

    def is_silent_exception(self, e: Exception):
        return isinstance(e, MqttError)


class MqttPlugin(BaseConnectionPluginConnectedTask[MqttConnection]):

    def __init__(self, task_name: str):
        super().__init__(self._mqtt_wrap_task, task_name)

    async def mqtt_task(self):
        raise NotImplementedError()

    async def _mqtt_wrap_task(self):

        connection = self.plugin_connection
        log = connection.log
        log.debug(f'{self.task.name} task start')
        try:
            with AsyncContext('MQTT'), connection.handle_exception(self.mqtt_task):
                await self.mqtt_task()
        finally:
            log.debug(f'{self.task.name} task stop')



from .wait_for_startlevel import WaitForStartlevelPlugin
from .load_items import LoadOpenhabItemsPlugin
from .events_sse import SseEventListenerPlugin
from .out import OUTGOING_PLUGIN, async_send_command, async_post_update, send_command, post_update
from .load_transformations import LoadTransformationsPlugin
from .ping import PingPlugin
from .wait_for_restore import WaitForPersistenceRestore
from .overview_things import ThingOverviewPlugin
from .plugin_things import TextualThingConfigPlugin
from .overview_broken_links import BrokenLinksPlugin



from typing import Optional, FrozenSet, Mapping, Union, TYPE_CHECKING

from HABApp.openhab.items.base_item import OpenhabItem, MetaData
from ..definitions import QuantityValue

if TYPE_CHECKING:
    Union = Union
    MetaData = MetaData
    FrozenSet = FrozenSet
    Mapping = Mapping


class NumberItem(OpenhabItem):
    """NumberItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Union[int, float] value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @property
    def unit(self) -> Optional[str]:
        """Return the item unit if it is a "Unit of Measurement" item else None"""
        if (unit := self.metadata.get('unit')) is None:
            return None
        return unit.value

    @staticmethod
    def _state_from_oh_str(state: str):
        try:
            return int(state)
        except ValueError:
            return float(state)

    def set_value(self, new_value) -> bool:

        if isinstance(new_value, QuantityValue):
            return super().set_value(new_value.value)

        return super().set_value(new_value)



from typing import Optional, Union, List, Any, Dict

from msgspec import Struct, field


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/StateOption.java
class StateOptionResp(Struct):
    value: str
    label: Optional[str] = None


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/StateDescription.java
class StateDescriptionResp(Struct, kw_only=True):
    minimum: Union[int, float, None] = None
    maximum: Union[int, float, None] = None
    step: Union[int, float, None] = None
    pattern: Optional[str] = None
    read_only: bool = field(name='readOnly')
    options: List[StateOptionResp]


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/types/CommandOption.java
class CommandOptionResp(Struct):
    command: str
    label: Optional[str] = None


class CommandDescriptionResp(Struct):
    command_options: List[CommandOptionResp] = field(name='commandOptions')


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/GroupFunctionDTO.java
class GroupFunctionResp(Struct):
    name: str
    params: List[str] = []


class ItemResp(Struct, kw_only=True):
    # ItemDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core/src/main/java/org/openhab/core/items/dto/ItemDTO.java
    type: str
    name: str
    label: Optional[str] = None
    category: Optional[str] = None
    tags: List[str]
    groups: List[str] = field(name='groupNames')

    # EnrichedItemDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/item/EnrichedItemDTO.java
    link: Optional[str] = None
    state: str
    transformed_state: Optional[str] = field(default=None, name='transformedState')
    state_description: Optional[StateDescriptionResp] = field(default=None, name='stateDescription')
    unit: Optional[str] = field(default=None, name='unitSymbol')
    command_description: Optional[CommandDescriptionResp] = field(default=None, name='commandDescription')
    metadata: Dict[str, Any] = {}
    editable: bool = True

    # EnrichedGroupItemDTO
    # https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest.core/src/main/java/org/openhab/core/io/rest/core/item/EnrichedGroupItemDTO.java
    members: List['ItemResp'] = []
    group_type: Optional[str] = field(default=None, name='groupType')
    group_function: Optional[GroupFunctionResp] = field(default=None, name='function')


class ShortItemResp(Struct):
    type: str
    name: str
    state: str



class Threshold:
    def __init__(self, threshold1, threshold2):
        """This is a simple Schmitt Trigger implementation.
        If the value is > upper_threshold is_higher will return true.
        The return value will stay true until the value goes below the lower threshold.

        :param threshold1:
        :param threshold2:
        """
        vals = sorted([threshold1, threshold2])
        self.lower_threshold = vals[0]
        self.upper_threshold = vals[1]
        assert self.lower_threshold < self.upper_threshold

        self.__threshold = self.upper_threshold

    def is_on(self, value):
        self.check_value(value)
        return self.__threshold == self.lower_threshold

    def is_off(self, value):
        return not self.is_on(value)

    @property
    def current_threshold(self):
        return self.__threshold

    def check_value(self, value):
        if value > self.upper_threshold:
            self.__threshold = self.lower_threshold

        if value < self.lower_threshold:
            self.__threshold = self.upper_threshold

        return self.__threshold

    def __lt__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.check_value(other) < other

    def __le__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.check_value(other) <= other

    def __ge__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.check_value(other) >= other

    def __gt__(self, other):
        if not isinstance(other, (int, float)):
            return NotImplemented

        return self.check_value(other) > other



from __future__ import annotations

import warnings
from datetime import datetime
from typing import Any, List
from urllib.parse import quote as quote_url

from HABApp.core.const.json import decode_struct
from HABApp.core.internals import ItemRegistryItem
from HABApp.openhab.definitions.rest import PersistenceServiceResp
from HABApp.openhab.definitions.rest import RootResp, SystemInfoRootResp, ItemResp, ShortItemResp, ItemHistoryResp, \
    ItemChannelLinkResp
from HABApp.openhab.definitions.rest import ThingResp
from HABApp.openhab.definitions.rest import TransformationResp
from HABApp.openhab.errors import (ThingNotFoundError, ItemNotFoundError, ItemNotEditableError,
                                   MetadataNotEditableError, ThingNotEditableError, TransformationsRequestError,
                                   PersistenceRequestError, LinkRequestError, LinkNotFoundError, LinkNotEditableError)
from . import convert_to_oh_type
from .handler import get, put, delete, post
from ...definitions.rest.habapp_data import get_api_vals, load_habapp_meta


# ----------------------------------------------------------------------------------------------------------------------
# root
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_root() -> RootResp | None:
    resp = await get('/rest/', log_404=False)
    if resp.status == 404 or resp.status == 500:
        return None

    # during startup, we sometimes get an empty response
    if not (b := await resp.read()):
        return None

    return decode_struct(b, type=RootResp)


# ----------------------------------------------------------------------------------------------------------------------
# uuid
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_uuid() -> str:
    resp = await get('/rest/uuid', log_404=False)
    return await resp.text(encoding='utf-8')


# ----------------------------------------------------------------------------------------------------------------------
# /systeminfo
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_system_info():
    resp = await get('/rest/systeminfo', log_404=False)
    if resp.status == 404 or resp.status == 500:
        return None

    # during startup, we sometimes get an empty response
    if not (b := await resp.read()):
        return None

    return decode_struct(b, type=SystemInfoRootResp).system_info


# ----------------------------------------------------------------------------------------------------------------------
# /items
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_items() -> list[ItemResp]:

    resp = await get('/rest/items', params={'metadata': '.+'})
    body = await resp.read()

    return decode_struct(body, type=List[ItemResp])


async def async_get_item(item: str | ItemRegistryItem) -> ItemResp | None:
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    resp = await get(f'/rest/items/{item:s}', log_404=False, params={'metadata': '.+'})
    if resp.status == 404:
        return None

    body = await resp.read()

    return decode_struct(body, type=ItemResp)


async def async_get_all_items_state() -> list[ShortItemResp]:
    resp = await get('/rest/items', params={'fields': 'name,state,type'})
    body = await resp.read()

    return decode_struct(body, type=List[ShortItemResp])


async def async_item_exists(item: str | ItemRegistryItem) -> bool:
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    ret = await get(f'/rest/items/{item:s}', log_404=False)
    return ret.status == 200


async def async_remove_item(item: str | ItemRegistryItem):
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    if (ret := await delete(f'/rest/items/{item:s}')) is None:
        return None

    if ret.status == 404:
        raise ItemNotFoundError.from_name(item)
    elif ret.status == 405:
        raise ItemNotEditableError.from_name(item)
    return ret.status < 300


async def async_create_item(item_type: str, name: str,
                            label: str | None = None, category: str | None = None,
                            tags: list[str] | None = None, groups: list[str] | None = None,
                            group_type: str | None = None,
                            group_function: str | None = None,
                            group_function_params: list[str] | None = None) -> bool:

    payload = {'type': item_type, 'name': name}
    if label:
        payload['label'] = label
    if category:
        payload['category'] = category
    if tags:
        payload['tags'] = tags
    if groups:
        payload['groupNames'] = groups  # CamelCase!

    # we create a group
    if group_type:
        payload['groupType'] = group_type   # CamelCase!
    if group_function:
        payload['function'] = {}
        payload['function']['name'] = group_function
        if group_function_params:
            payload['function']['params'] = group_function_params

    if (ret := await put(f'/rest/items/{name:s}', json=payload)) is None:
        return False

    if ret.status == 404:
        raise ItemNotFoundError.from_name(name)
    elif ret.status == 405:
        raise ItemNotEditableError.from_name(name)
    return ret.status < 300


async def async_remove_metadata(item: str | ItemRegistryItem, namespace: str):
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    if (ret := await delete(f'/rest/items/{item:s}/metadata/{namespace:s}')) is None:
        return False

    if ret.status == 404:
        raise ItemNotFoundError.from_name(item)
    elif ret.status == 405:
        raise MetadataNotEditableError.create_text(item, namespace)
    return ret.status < 300


async def async_set_metadata(item: str | ItemRegistryItem, namespace: str, value: str, config: dict):
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    payload = {
        'value': value,
        'config': config
    }
    ret = await put(f'/rest/items/{item:s}/metadata/{namespace:s}', json=payload)
    if ret is None:
        return False

    if ret.status == 404:
        raise ItemNotFoundError.from_name(item)
    elif ret.status == 405:
        raise MetadataNotEditableError.create_text(item, namespace)
    return ret.status < 300


# ----------------------------------------------------------------------------------------------------------------------
# /things
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_things() -> list[ThingResp]:
    resp = await get('/rest/things')
    body = await resp.read()

    return decode_struct(body, type=List[ThingResp])


async def async_get_thing(thing: str | ItemRegistryItem) -> ThingResp:
    # noinspection PyProtectedMember
    thing = thing if isinstance(thing, str) else thing._name

    resp = await get(f'/rest/things/{thing:s}')
    if resp.status >= 300:
        raise ThingNotFoundError.from_uid(thing)

    body = await resp.read()
    return decode_struct(body, type=ThingResp)


async def async_set_thing_cfg(thing: str | ItemRegistryItem, cfg: dict[str, Any]):
    # noinspection PyProtectedMember
    thing = thing if isinstance(thing, str) else thing._name

    if (ret := await put(f'/rest/things/{thing:s}/config', json=cfg)) is None:
        return None

    if ret.status == 404:
        raise ThingNotFoundError.from_uid(thing)
    elif ret.status == 409:
        raise ThingNotEditableError.from_uid(thing)
    elif ret.status >= 300:
        raise ValueError('Something went wrong')

    return ret.status


async def async_set_thing_enabled(thing: str | ItemRegistryItem, enabled: bool):
    # noinspection PyProtectedMember
    thing = thing if isinstance(thing, str) else thing._name

    if (ret := await put(f'/rest/things/{thing:s}/enable', data='true' if enabled else 'false')) is None:
        return None

    if ret.status == 404:
        raise ThingNotFoundError.from_uid(thing)
    elif ret.status == 409:
        raise ThingNotEditableError.from_uid(thing)
    elif ret.status >= 300:
        raise ValueError('Something went wrong')

    return ret.status


# ----------------------------------------------------------------------------------------------------------------------
# /links
# ----------------------------------------------------------------------------------------------------------------------
async def async_purge_links():
    resp = await post('/rest/purge')
    if resp.status != 200:
        raise LinkRequestError('Unexpected error')


async def async_remove_obj_links(name: str | ItemRegistryItem) -> bool:
    """Remove links from an item or a thing

    :param name: name of thing or item
    """
    # noinspection PyProtectedMember
    name = name if isinstance(name, str) else name._name

    resp = await delete(f'/rest/links/{name:s}')
    if resp.status >= 300:
        raise LinkRequestError()

    return True


async def async_get_links() -> list[ItemChannelLinkResp]:

    resp = await get('/rest/links')
    if resp.status != 200:
        raise LinkRequestError('Unexpected error')

    body = await resp.read()
    return decode_struct(body, type=List[ItemChannelLinkResp])


def __get_item_link_url(item: str | ItemRegistryItem, channel: str) -> str:
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    # rest/links/ endpoint needs the channel to be url encoded
    # (AAAA:BBBB:CCCC:0#NAME -> AAAA%3ABBBB%3ACCCC%3A0%23NAME)
    # otherwise the REST-api returns HTTP-Status 500 InternalServerError
    return '/rest/links/' + quote_url(f"{item:s}/{channel:s}")


async def async_get_link(item: str | ItemRegistryItem, channel: str) -> ItemChannelLinkResp:

    resp = await get(__get_item_link_url(item, channel), log_404=False)
    if resp.status == 200:
        body = await resp.read()
        return decode_struct(body, type=ItemChannelLinkResp)

    if resp.status == 404:
        raise LinkNotFoundError.from_names(item, channel)
    raise LinkRequestError('Unexpected error')


async def async_create_link(
        item: str | ItemRegistryItem, channel: str, configuration: dict[str, Any] | None = None) -> bool:
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    # if the passed item doesn't exist OpenHAB creates a new empty item
    # this is undesired and why we raise an Exception
    if not await async_item_exists(item):
        raise ItemNotFoundError.from_name(item)

    json = {'itemName': item, 'channelUID': channel}
    if configuration is not None:
        json['configuration'] = configuration

    if (resp := await put(__get_item_link_url(item, channel), json=json)) is None:
        return False

    if resp.status == 200:
        return True

    if resp.status == 405:
        LinkNotEditableError.from_names(item, channel)
    raise LinkRequestError('Unexpected error')


async def async_remove_link(item: str | ItemRegistryItem, channel: str):

    if (resp := await delete(__get_item_link_url(item, channel), log_404=False)) is None:
        return None
    if resp.status == 200:
        return None

    if resp.status == 404:
        raise LinkNotFoundError.from_names(item, channel)
    if resp.status == 405:
        LinkNotEditableError.from_names(item, channel)
    raise LinkRequestError('Unexpected error')


# ----------------------------------------------------------------------------------------------------------------------
# /transformations
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_transformations() -> list[TransformationResp]:
    resp = await get('/rest/transformations')
    if resp.status >= 300:
        raise TransformationsRequestError()

    body = await resp.read()
    return decode_struct(body, type=List[TransformationResp])


# ----------------------------------------------------------------------------------------------------------------------
# /persistence
# ----------------------------------------------------------------------------------------------------------------------
async def async_get_persistence_services() -> list[PersistenceServiceResp]:
    resp = await get('/rest/persistence')
    if resp.status >= 300:
        raise PersistenceRequestError()

    body = await resp.read()
    return decode_struct(body, type=List[PersistenceServiceResp])


async def async_get_persistence_data(item: str | ItemRegistryItem, persistence: str | None,
                                     start_time: datetime | None,
                                     end_time: datetime | None) -> ItemHistoryResp:
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    params = {}
    if persistence:
        params['serviceId'] = persistence
    if start_time is not None:
        params['starttime'] = convert_to_oh_type(start_time)
    if end_time is not None:
        params['endtime'] = convert_to_oh_type(end_time)
    if not params:
        params = None

    resp = await get(f'/rest/persistence/items/{item:s}', params=params)
    if resp.status >= 300:
        raise PersistenceRequestError()

    body = await resp.read()
    return decode_struct(body, type=ItemHistoryResp)


async def async_set_persistence_data(item: str | ItemRegistryItem, persistence: str | None,
                                     time: datetime, state: Any):
    # noinspection PyProtectedMember
    item = item if isinstance(item, str) else item._name

    # This does only work for some persistence services (as of OH 3.4)
    warnings.warn(f'{async_set_persistence_data.__name__} calls a part of the openHAB API which is buggy!',
                  category=ResourceWarning)

    params = {
        'time': convert_to_oh_type(time),
        'state': convert_to_oh_type(state),
    }
    if persistence is not None:
        params['serviceId'] = persistence

    ret = await put(f'/rest/persistence/items/{item:s}', params=params)
    if ret.status >= 300:
        return None
    else:
        # I would have expected the endpoint to return a valid json, but instead it returns nothing
        # return await ret.json(loads=load_json, encoding='utf-8')
        return None


# ---------------------------------------------------------------------------------------------------------------------
# Funcs for handling HABApp Metadata
# ---------------------------------------------------------------------------------------------------------------------
async def async_remove_habapp_metadata(item: str):
    return await async_remove_metadata(item, 'HABApp')


async def async_set_habapp_metadata(item: str, obj):
    val, cfg = get_api_vals(obj)
    return await async_set_metadata(item, 'HABApp', val, cfg)


async def async_get_item_with_habapp_meta(item: str) -> ItemResp:
    if (data := await async_get_item(item)) is None:
        raise ItemNotFoundError.from_name(item)
    return load_habapp_meta(data)



import logging
# noinspection PyProtectedMember
from sys import _getframe as sys_get_frame
from types import FrameType
from typing import TYPE_CHECKING, Any, Callable, Final, Optional

if TYPE_CHECKING:
    import HABApp
    import HABApp.rule_manager

_NAME: Final = '__HABAPP__HOOK__'


log = logging.getLogger('HABApp.Rule')


class HABAppRuleHook:

    def __init__(self,
                 cb_register_rule: Callable[['HABApp.rule.Rule'], Any],
                 cb_suggest_name: Callable[['HABApp.rule.Rule'], str],
                 runtime: 'HABApp.runtime.Runtime', rule_file: 'HABApp.rule_manager.RuleFile'):
        # callbacks
        self._cb_register: Final = cb_register_rule
        self._cb_suggest_name: Final = cb_suggest_name

        # runtime objs
        self.runtime: Final = runtime
        self.rule_file: Final = rule_file

        self.closed = False

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.closed = True

    def register_rule(self, rule: 'HABApp.rule.Rule'):
        if self.closed:
            # if we keep adding rules dynamically they will always get attached to the file and never unloaded
            log.warning(f'Added another rule of type {rule.__class__.__name__:s} '
                        'but file load has already been completed!')
        return self._cb_register(rule)

    def suggest_rule_name(self, rule: 'HABApp.rule.Rule') -> str:
        return self._cb_suggest_name(rule)

    def in_dict(self, obj: Optional[dict] = None) -> dict:
        if obj is None:
            obj = {}
        obj[_NAME] = self
        return obj


def get_rule_hook() -> HABAppRuleHook:

    # noinspection PyUnresolvedReferences
    frame: Optional[FrameType] = sys_get_frame(1)

    while frame is not None:
        _globals = frame.f_globals

        if (hook := _globals.get(_NAME)) is not None:
            assert isinstance(hook, HABAppRuleHook)
            return hook

        frame = frame.f_back

    raise RuntimeError('HABApp rule files are not meant to be executed directly! '
                       'Put the file in the HABApp "rule" folder and HABApp will load it automatically.')



from datetime import datetime
from typing import TYPE_CHECKING, Optional, FrozenSet, Mapping

from HABApp.core.const.const import PYTHON_311
from HABApp.openhab.items.base_item import OpenhabItem, MetaData

if TYPE_CHECKING:
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


class DatetimeItem(OpenhabItem):
    """DateTimeItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar datetime value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        # see implementation im map_values.py
        if PYTHON_311:
            dt = datetime.fromisoformat(state)
        else:
            pos_dot = state.find('.')
            if (pos_plus := state.rfind('+')) == -1:
                pos_plus = state.rfind('-')
            if pos_plus - pos_dot > 6:
                state = state[:pos_dot + 7] + state[pos_plus:]
            dt = datetime.strptime(state, '%Y-%m-%dT%H:%M:%S.%f%z')

        # all datetime objs from openHAB have a timezone set so we can't easily compare them
        # --> TypeError: can't compare offset-naive and offset-aware datetime
        dt = dt.astimezone(tz=None)  # Changes datetime object so it uses system timezone
        value = dt.replace(tzinfo=None)  # Removes timezone awareness
        return value



from typing import TYPE_CHECKING, Optional, FrozenSet, Mapping

from HABApp.openhab.items.base_item import OpenhabItem, MetaData

if TYPE_CHECKING:
    MetaData = MetaData
    Optional = Optional
    FrozenSet = FrozenSet,
    Mapping = Mapping


class StringItem(OpenhabItem):
    """StringItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        return state


class PlayerItem(OpenhabItem):
    """PlayerItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar str value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        return state



from typing import List

from msgspec import Struct


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest/src/main/java/org/openhab/core/io/rest/internal/resources/beans/RootBean.java


class RuntimeResp(Struct, rename='camel'):
    version: str
    build_string: str


class LinkResp(Struct):
    type: str
    url: str


class RootResp(Struct, rename='camel'):
    version: str
    locale: str
    measurement_system: str
    runtime_info: RuntimeResp
    links: List[LinkResp]



import platform
import sys
from HABApp.__version__ import __version__


def get_debug_info() -> str:

    info = {
        'HABApp': __version__,
        'Platform': platform.platform(),
        'Machine': platform.machine(),
        'Python': sys.version,
    }

    indent = max(map(lambda x: len(x), info))
    ret = '\n'.join('{:{indent}s}: {:s}'.format(k, str(v).replace('\n', ''), indent=indent) for k, v in info.items())

    try:
        import pkg_resources
        installed_packages = {p.key: p.version for p in sorted(pkg_resources.working_set, key=lambda x: x.key)}

        indent = max(map(len, installed_packages.keys()), default=1) + 2
        table = '\n'.join(f'{k:{indent}s}: {v}' for k, v in installed_packages.items())

        if installed_packages:
            ret += f'\n\nInstalled Packages\n{"-" * 80}\n{table}'

    except Exception as e:
        ret += f'\n\nCould not get installed Packages!\nError: {str(e)}'

    return ret


def print_debug_info():
    print(f'Debug information\n{"-" * 80}')
    print(get_debug_info())



from __future__ import annotations

import inspect

import eascheduler.scheduler_view
from HABApp.core.internals import uses_item_registry
from HABApp.core.items import BaseValueItem
from HABApp.core.wrapper import ignore_exception
from HABApp.openhab.items import OpenhabItem
from eascheduler.const import FAR_FUTURE
from eascheduler.jobs import (
    CountdownJob as CountdownJobBase, DawnJob as DawnJobBase, DayOfWeekJob as DayOfWeekJobBase,
    DuskJob as DuskJobBase, OneTimeJob as OneTimeJobBase, ReoccurringJob as ReoccurringJobBase,
    SunriseJob as SunriseJobBase, SunsetJob as SunsetJobBase
)
from eascheduler.jobs.job_base import ScheduledJobBase

Items = uses_item_registry()


class ItemBoundJobMixin:
    def __init__(self):
        super().__init__()
        self._item: BaseValueItem | None = None

    @ignore_exception
    def _timestamp_to_item(self, job: ScheduledJobBase):
        if self._item is None:
            return None

        if self._next_run >= FAR_FUTURE:
            next_run = None
        else:
            next_run = self.get_next_run()

        if isinstance(self._item, OpenhabItem):
            self._item.oh_post_update(next_run)
        else:
            self._item.post_value(next_run)

        # if the job has been canceled we clean this up, too
        if self._parent is None:
            self._item = None
            self._next_run_callback = None

    def to_item(self, item: str | BaseValueItem | None):
        """Sends the next execution (date)time to an item. Sends ``None`` if the job is not scheduled.

        :param item: item name or item, None to disable
        """
        if item is None:
            self._item = None
            self._next_run_callback = None
        else:
            self._item = Items.get_item(item) if not isinstance(item, BaseValueItem) else item
            self._next_run_callback = self._timestamp_to_item
            # Update the item with the current timestamp
            self._timestamp_to_item(self)

    # I am not sure about this operator - it seems like a nice idea but doesn't work well
    # with the code inspections from the IDE
    #
    # def __gt__(self, other):
    #     if isinstance(other, (str, BaseValueItem, None)):
    #         self.to_item(other)
    #         return self
    #     return NotImplemented


class CountdownJob(CountdownJobBase, ItemBoundJobMixin):
    pass


class DawnJob(DawnJobBase, ItemBoundJobMixin):
    pass


class DayOfWeekJob(DayOfWeekJobBase, ItemBoundJobMixin):
    pass


class DuskJob(DuskJobBase, ItemBoundJobMixin):
    pass


class OneTimeJob(OneTimeJobBase, ItemBoundJobMixin):
    pass


class ReoccurringJob(ReoccurringJobBase, ItemBoundJobMixin):
    pass


class SunriseJob(SunriseJobBase, ItemBoundJobMixin):
    pass


class SunsetJob(SunsetJobBase, ItemBoundJobMixin):
    pass


# This is a very dirty hack - I really should come up with something different
def replace_jobs():
    g = globals()
    module = eascheduler.scheduler_view
    for name, obj in inspect.getmembers(module):
        if not name.endswith('Job'):
            continue

        assert obj is g[name + 'Base']
        setattr(module, name, g[name])


replace_jobs()



from __future__ import annotations

from aiomqtt import Client, TLSParameters

from HABApp.config import CONFIG
from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.connections._definitions import CONNECTION_HANDLER_NAME

from HABApp.core.internals import uses_post_event, uses_get_item, uses_item_registry
from HABApp.mqtt.connection.connection import CONTEXT_TYPE, MqttConnection

post_event = uses_post_event()
get_item = uses_get_item()
Items = uses_item_registry()


class ConnectionHandler(BaseConnectionPlugin[MqttConnection]):
    def __init__(self):
        super().__init__(name=CONNECTION_HANDLER_NAME)

    async def on_setup(self, connection: MqttConnection):
        log = connection.log
        config = CONFIG.mqtt.connection

        if not config.host:
            log.info('MQTT disabled')
            connection.status_from_setup_to_disabled()
            return None

        tls_insecure: bool | None = None
        tls_ca_cert: str | None = None
        if tls_enabled := config.tls.enabled:
            log.debug("TLS enabled")

            # add option to specify tls certificate
            ca_cert = config.tls.ca_cert
            if ca_cert is not None and ca_cert.name:
                if not ca_cert.is_file():
                    log.error(f'Ca cert file does not exist: {ca_cert}')
                    # don't connect without the properly set certificate
                    connection.set_error()
                    return None

                log.debug(f"CA cert path: {ca_cert}")
                tls_ca_cert = str(ca_cert)

            # we can only set tls_insecure if we have a tls connection
            if config.tls.insecure:
                log.warning('Verification of server hostname in server certificate disabled!')
                log.warning('Use this only for testing, not for a real system!')
                tls_insecure = True

        connection.context = Client(
            hostname=config.host, port=config.port,
            username=config.user if config.user else None, password=config.password if config.password else None,
            client_id=config.client_id,

            tls_insecure=tls_insecure,
            tls_params=None if not tls_enabled else TLSParameters(ca_certs=tls_ca_cert),

            # clean_session=False
        )

    # noinspection PyProtectedMember
    async def on_connecting(self, connection: MqttConnection, context: CONTEXT_TYPE):
        assert context is not None

        connection.log.info(f'Connecting to {context._hostname}:{context._port}')
        await context.__aenter__()
        connection.log.info('Connection successful')

    async def on_disconnected(self, connection: MqttConnection, context: CONTEXT_TYPE):
        assert context is not None

        connection.log.info('Disconnected')
        # remove this check when https://github.com/sbtinstruments/aiomqtt/pull/249 gets merged
        if not context._lock.locked():
            await context._lock.acquire()
        await context.__aexit__(None, None, None)


CONNECTION_HANDLER = ConnectionHandler()



from pathlib import Path
from typing import Dict
from typing import List, Type

import HABApp
from HABApp.core.const.topics import TOPIC_FILES as T_FILES
from HABApp.core.events.habapp_events import RequestFileUnloadEvent, RequestFileLoadEvent
from ..watcher import AggregatingAsyncEventHandler
from HABApp.core.internals import uses_post_event

FOLDERS: Dict[str, 'ConfiguredFolder'] = {}

post_event = uses_post_event()


async def _generate_file_events(files: List[Path]):
    for file in files:
        name = get_name(file)
        post_event(T_FILES, RequestFileLoadEvent(name) if file.is_file() else RequestFileUnloadEvent(name))


class ConfiguredFolder:
    def __init__(self, prefix: str, folder: Path, priority: int):
        self.prefix = prefix
        self.folder = folder
        self.priority: int = priority   # priority determines the order how the files will be loaded

    def add_watch(self, file_ending: str, watch_subfolders: bool = True) -> AggregatingAsyncEventHandler:
        filter = HABApp.core.files.watcher.FileEndingFilter(file_ending)
        handler = AggregatingAsyncEventHandler(self.folder, _generate_file_events, filter, watch_subfolders)
        HABApp.core.files.watcher.add_folder_watch(handler)
        return handler

    def add_file_type(self, cls: Type['HABApp.core.files.file.HABAppFile']):
        HABApp.core.files.file.register_file_type(self.prefix, cls)


def get_prefixes() -> List[str]:
    return list(map(lambda x: x.prefix, sorted(FOLDERS.values(), key=lambda x: x.priority, reverse=True)))


def add_folder(prefix: str, folder: Path, priority: int) -> ConfiguredFolder:
    """Make a folder known

    :param prefix: HABApp file name prefix
    :param folder: Folder path
    :param priority: Priority (used to determine the load order)
    :return: ConfiguredFolder
    """
    assert prefix and prefix.endswith('/')
    for obj in FOLDERS.values():
        assert obj.priority != priority
    FOLDERS[prefix] = c = ConfiguredFolder(prefix, folder, priority)
    return c


def get_name(path: Path) -> str:
    path_str = path.as_posix()
    for prefix, cfg in sorted(FOLDERS.items(), key=lambda x: len(x[0]), reverse=True):
        folder = cfg.folder.as_posix()
        if path_str.startswith(folder):
            return prefix + path_str[len(folder) + 1:]

    raise ValueError(f'Path "{path_str}" is not part of the configured folders!')


def get_path(name: str) -> Path:
    for prefix, obj in FOLDERS.items():
        if name.startswith(prefix):
            return obj.folder / name[len(prefix):]

    raise ValueError(f'Prefix not found for "{name}"!')



from typing import Dict, Union, Tuple, Any, Final

from javaproperties import loads as load_map_file

from HABApp.openhab.errors import MapTransformationNotFound
from HABApp.openhab.transformations._map.classes import MapTransformation, MapTransformationWithDefault
from HABApp.openhab.transformations.base import TransformationRegistryBase, TransformationFactoryBase, log


class MapTransformationRegistry(TransformationRegistryBase):

    def __init__(self, name: str):
        super().__init__(name)
        self.objs: Dict[str, Tuple[dict, Any]] = {}

    def get(self, name: str) -> Union[MapTransformation, MapTransformationWithDefault]:
        try:
            data, default = self.objs[name]
        except KeyError:
            raise MapTransformationNotFound(f'Map transformation "{name:s}" not found!') from None

        if default:
            return MapTransformationWithDefault(data, name=name, default=default)
        else:
            return MapTransformation(data, name=name)

    def set(self, name: str, configuration: Dict[str, str]):
        data = load_map_file(configuration['function'])
        if not data:
            log.warning(f'Map transformation "{name:s}" is empty -> skipped!')
            return None

        default = data.pop('', None)
        map_keys = all(k.isdecimal() for k in data)
        map_values = all(v.isdecimal() for v in data.values())
        if map_values and default is not None and default.isdecimal():
            default = int(default)

        obj = {}
        for k, v in data.items():
            key = k if not map_keys else int(k)
            value = v if not map_values else int(v)
            obj[key] = value

        self.objs[name] = obj, default


MAP_REGISTRY: Final = MapTransformationRegistry('map')


class MapTransformationFactory(TransformationFactoryBase[Dict[Union[str, int], Union[str, int]]]):
    pass


MAP_FACTORY = MapTransformationFactory(MAP_REGISTRY)



from HABApp.core.internals.proxy.proxy_obj import create_proxy, replace_proxies
from typing import Callable, TYPE_CHECKING, Any

if TYPE_CHECKING:
    import HABApp


def uses_post_event() -> Callable[[str, Any], None]:
    return create_proxy(uses_post_event)


def uses_event_bus() -> 'HABApp.core.internals.EventBus':
    return create_proxy(uses_event_bus)


def uses_get_item() -> Callable[[str], 'HABApp.core.internals.item_registry.item_registry.ItemRegistryItem']:
    return create_proxy(uses_get_item)


def uses_item_registry() -> 'HABApp.core.internals.ItemRegistry':
    return create_proxy(uses_item_registry)


def setup_internals(ir: 'HABApp.core.internals.ItemRegistry',
                    eb: 'HABApp.core.internals.EventBus', final=True):
    """Replace the proxy objects with the real thing"""
    replacements = {
        uses_item_registry: ir, uses_get_item: ir.get_item,
        uses_event_bus: eb, uses_post_event: eb.post_event,
    }
    return replace_proxies(replacements, final=final)



from __future__ import annotations

from datetime import datetime
from typing import Any

from HABApp.core.items import BaseValueItem
from HABApp.core.types import RGB, HSB


def convert_to_oh_type(obj: Any) -> str:
    if isinstance(obj, (str, int, float, bool)):
        return str(obj)

    if isinstance(obj, datetime):
        # Add timezone (if not yet defined) to string, then remote anything below ms.
        # 2018-11-19T09:47:38.284000+0100 -> 2018-11-19T09:47:38.284+0100
        out = obj.astimezone(None).strftime('%Y-%m-%dT%H:%M:%S.%f%z')
        return out

    if isinstance(obj, (set, list, tuple, frozenset)):
        return ','.join(map(str, obj))

    if obj is None:
        return 'NULL'

    if isinstance(obj, RGB):
        obj = obj.to_hsb()

    if isinstance(obj, HSB):
        # noinspection PyProtectedMember
        return f'{obj._hue:.2f},{obj._saturation:.2f},{obj._brightness:.2f}'

    if isinstance(obj, BaseValueItem):
        raise ValueError()

    return str(obj)


def map_null_str(value: str) -> str | None:
    if value == 'NULL' or value == 'UNDEF':
        return None
    return value



from __future__ import annotations

from enum import Enum, auto


class FileState(Enum):
    LOADED = auto()
    FAILED = auto()

    DEPENDENCIES_OK = auto()
    DEPENDENCIES_MISSING = auto()
    DEPENDENCIES_ERROR = auto()

    PROPERTIES_INVALID = auto()  # Properties could not be parsed

    # initial and last state
    PENDING = auto()
    REMOVED = auto()

    def __str__(self):
        return str(self.name)



from __future__ import annotations

import logging
from pathlib import Path

from pydantic import ValidationError

from HABApp.core.files.file import HABAppFile, FileState
from HABApp.core.files.file.properties import get_properties, FileProperties
from HABApp.core.logger import HABAppError

FILE_TYPES: dict[str, type[HABAppFile]] = {}


log = logging.getLogger('HABApp.files')


def register_file_type(prefix: str, cls: type[HABAppFile]):
    assert prefix not in FILE_TYPES

    assert cls.LOGGER
    assert cls.LOAD_FUNC
    assert cls.UNLOAD_FUNC

    FILE_TYPES[prefix] = cls


def create_file(name: str, path: Path) -> HABAppFile:
    for prefix, cls in FILE_TYPES.items():
        if name.startswith(prefix):
            break
    else:
        raise ValueError(f'Unknown file type for "{name}"!')

    with path.open('r', encoding='utf-8') as f:
        txt = f.read(10 * 1024)

    validation_error = True

    try:
        properties = get_properties(txt)
        validation_error = False
    except ValidationError as e:
        logger = HABAppError(log)
        logger.add(f'Error while parsing properties for {name:s}:')
        for line in str(e).splitlines()[1:]:
            logger.add(f'  {line:s}')
        logger.dump()

        properties = FileProperties()

    obj = cls(name, path, properties)
    if validation_error:
        obj.set_state(FileState.PROPERTIES_INVALID)

    return obj



from __future__ import annotations

from HABApp.core.connections import BaseConnectionPlugin
from HABApp.core.internals import uses_item_registry
from HABApp.core.wrapper import ExceptionToHABApp
from HABApp.openhab.connection.connection import OpenhabConnection, OpenhabContext
from HABApp.openhab.connection.handler.func_async import async_get_transformations
from HABApp.openhab.transformations._map import MAP_REGISTRY
from HABApp.openhab.transformations.base import log, TransformationRegistryBase

Items = uses_item_registry()


class LoadTransformationsPlugin(BaseConnectionPlugin[OpenhabConnection]):

    async def on_connected(self, context: OpenhabContext):
        if context.is_oh3:
            log.info('Transformations are not supported on openHAB 3')
            return None

        exception_handler = ExceptionToHABApp(logger=log)

        log.debug('Requesting transformations')
        objs = await async_get_transformations()
        transformation_count = len(objs)
        log.debug(f'Got response with {transformation_count} transformation{"" if transformation_count == 1 else ""}')

        registries: dict[str, TransformationRegistryBase] = {
            MAP_REGISTRY.name: MAP_REGISTRY
        }

        for reg in registries.values():
            reg.clear()

        for obj in objs:
            with exception_handler:
                if reg := registries.get(obj.type):
                    reg.set(obj.uid, obj.configuration)

        if not any(r.objs for r in registries.values()):
            log.info('No transformations available')
        else:
            log.info('Transformations:')
            for name, reg in registries.items():
                log.info(f'  {name.title()}: {", ".join(reg.available())}')



from asyncio import Queue
from typing import Optional, Union

from HABApp.config import CONFIG
from HABApp.config.models.mqtt import QOS
from HABApp.core.asyncio import run_func_from_async
from HABApp.core.const.json import dump_json
from HABApp.core.internals import ItemRegistryItem
from HABApp.mqtt.connection.connection import MqttPlugin


class PublishHandler(MqttPlugin):
    def __init__(self):
        super().__init__(task_name='MqttPublish')

    async def mqtt_task(self):
        connection = self.plugin_connection
        with connection.handle_exception(self.mqtt_task):
            client = self.plugin_connection.context
            assert client is not None

            cfg = CONFIG.mqtt.publish
            queue = QUEUE
            assert queue is not None

            # worker to publish things
            while True:
                topic, value, qos, retain = await queue.get()
                if qos is None:
                    qos = cfg.qos
                if retain is None:
                    retain = cfg.retain

                await client.publish(topic, value, qos, retain)
                queue.task_done()

    async def on_connected(self):
        global QUEUE

        QUEUE = Queue()
        await super().on_connected()

    async def on_disconnected(self):
        global QUEUE

        await super().on_disconnected()
        QUEUE = None


QUEUE: Optional[Queue] = Queue()


PUBLISH_HANDLER = PublishHandler()


def async_publish(topic: Union[str, ItemRegistryItem], payload, qos: Optional[QOS] = None,
                  retain: Optional[bool] = None):
    """
    Publish a value under a certain topic.
    If qos and/or retain is not set the value from the configuration file will be used.

    :param topic: MQTT topic or item
    :param payload: MQTT Payload
    :param qos: QoS, can be 0, 1 or 2. If not specified the value from configuration file will be used.
    :param retain: retain message. If not specified the value from configuration file will be used.
    """
    if (queue := QUEUE) is None:
        return None

    if isinstance(topic, ItemRegistryItem):
        topic = topic.name

    # convert these to string
    if isinstance(payload, (dict, list, set, frozenset)):
        payload = dump_json(payload)

    queue.put_nowait((topic, payload, qos, retain))


def publish(topic: Union[str, ItemRegistryItem], payload, qos: Optional[QOS] = None, retain: Optional[bool] = None):
    """
    Publish a value under a certain topic.
    If qos and/or retain is not set the value from the configuration file will be used.

    :param topic: MQTT topic or item
    :param payload: MQTT Payload
    :param qos: QoS, can be 0, 1 or 2. If not specified the value from configuration file will be used.
    :param retain: retain message. If not specified the value from configuration file will be used.
    """
    run_func_from_async(topic, payload, qos, retain)



import random
import time
from collections import deque
from threading import Lock

import HABApp
from HABApp.core.events import ValueUpdateEvent, ValueUpdateEventFilter
from .bench_base import BenchBaseRule
from .bench_times import BenchContainer, BenchTime

LOCK = Lock()


class OpenhabBenchRule(BenchBaseRule):
    BENCH_TYPE = 'openHAB'

    RTT_BENCH_MAX = 15

    def __init__(self):
        super().__init__()

        self.name_list = [f'BenchItem{k}' for k in range(300)]

        self.time_sent = 0.0
        self.bench_started = 0.0
        self.bench_times_container = BenchContainer()
        self.bench_times: BenchTime = None

        self.item_values = deque()
        self.item_name = ''

        self.load_listener = []

    def cleanup(self):
        self.stop_load()

        all_items = set(HABApp.core.Items.get_item_names())
        to_rem = set(self.name_list) & all_items

        if not to_rem:
            return None

        print('Cleanup ... ', end='')
        for name in to_rem:
            self.oh.remove_item(name)
        print('complete')

    def set_up(self):
        self.cleanup()

    def tear_down(self):
        self.cleanup()

    def run_bench(self):
        # These are the benchmarks
        self.bench_item_create()
        self.bench_rtt_time()

    def bench_item_create(self):
        print('Bench item operations ', end='')

        max_duration = 10   # how long should each bench take

        times = BenchContainer()

        start_bench = time.time()
        b = times.create('create item')
        for k in self.name_list:
            start = time.time()
            self.openhab.create_item('Number', k, label='MyLabel')
            b.times.append(time.time() - start)

            # limit bench time on weak devices
            if time.time() - start_bench > max_duration:
                break

        time.sleep(0.2)

        print('.', end='')
        start_bench = time.time()
        b = times.create('update item')
        for k in self.name_list:
            start = time.time()
            self.openhab.create_item('Number', k, label='New Label')
            b.times.append(time.time() - start)

            # limit bench time on weak devices
            if time.time() - start_bench > max_duration:
                break

        time.sleep(0.2)

        print('.', end='')
        start_bench = time.time()
        b = times.create('delete item')
        for k in self.name_list:
            start = time.time()
            self.openhab.remove_item(k)
            b.times.append(time.time() - start)

            # limit bench time on weak devices
            if time.time() - start_bench > max_duration:
                break

        print('. done!\n')
        times.show()

    def bench_rtt_time(self):
        print('Bench item state update ', end='')
        self.bench_times_container = BenchContainer()

        self.run_rtt('rtt idle')
        self.run_rtt('async rtt idle', do_async=True)

        self.start_load()
        self.run_rtt('rtt load (+10x)')
        self.run_rtt('async rtt load (+10x)', do_async=True)
        self.stop_load()

        print(' done!\n')
        self.bench_times_container.show()

    def start_load(self):

        for i in range(10, 20):
            def load_cb(event, item=self.name_list[i]):
                self.openhab.post_update(item, str(random.randint(0, 99999999)))

            self.openhab.create_item('String', self.name_list[i], label='MyLabel')
            listener = self.listen_event(self.name_list[i], load_cb, ValueUpdateEventFilter())

            self.load_listener.append(listener)
            self.openhab.post_update(self.name_list[i], str(random.randint(0, 99999999)))

    def stop_load(self):
        for list in self.load_listener:
            list.cancel()
        self.load_listener.clear()

    def run_rtt(self, test_name, do_async=False):
        self.item_name = self.name_list[0]
        self.openhab.create_item('String', self.item_name, label='MyLabel')

        for i in range(3000):
            self.item_values.append(str(random.randint(0, 99999999)))

        listener = self.listen_event(
            self.item_name, self.proceed_item_val if not do_async else self.a_proceed_item_val, ValueUpdateEventFilter()
        )

        self.bench_times = self.bench_times_container.create(test_name)

        self.bench_started = time.time()
        self.time_sent = time.time()
        self.openhab.post_update(self.item_name, self.item_values[0])

        self.run.soon(LOCK.acquire)
        time.sleep(1)
        LOCK.acquire(True, OpenhabBenchRule.RTT_BENCH_MAX)

        listener.cancel()
        if LOCK.locked():
            LOCK.release()

        print('.', end='')

    def proceed_item_val(self, event: ValueUpdateEvent):
        if event.value != self.item_values[0]:
            return None

        self.bench_times.times.append(time.time() - self.time_sent)

        # Time up -> stop benchmark
        if time.time() - self.bench_started > OpenhabBenchRule.RTT_BENCH_MAX:
            LOCK.release()
            return None

        # No items left -> stop benchmark
        try:
            self.item_values.popleft()
        except IndexError:
            LOCK.release()
            return None

        self.time_sent = time.time()
        self.openhab.post_update(self.item_name, self.item_values[0])

    async def a_proceed_item_val(self, event: ValueUpdateEvent):
        self.proceed_item_val(event)



from pydantic import Field, conint

from easyconfig import BaseModel


class ThreadPoolConfig(BaseModel):
    enabled: bool = True
    """When the thread pool is disabled HABApp will become an asyncio application.
    Use only if you have experience developing asyncio applications!
    If the thread pool is disabled using blocking calls in functions can and will break HABApp"""

    threads: conint(ge=1, le=16) = 10
    """Amount of threads to use for the executor"""


class LoggingConfig(BaseModel):
    use_buffer: bool = Field(True, alias='use buffer')
    """Automatically inject a buffer for the event log"""

    flush_every: float = Field(0.5, alias='flush every', ge=0.1)
    """Wait time in seconds before the buffer gets flushed again when it was empty"""


class HABAppConfig(BaseModel):
    """HABApp internal configuration. Only change values if you know what you are doing!"""

    logging: LoggingConfig = Field(default_factory=LoggingConfig)
    thread_pool: ThreadPoolConfig = Field(default_factory=ThreadPoolConfig, alias='thread pool')



from inspect import getmembers, isclass
from pathlib import Path

from pydantic import BaseModel

import HABApp.config.models
from HABApp.config import CONFIG
from easyconfig import yaml


def test_sample_yaml(pytestconfig):
    file = pytestconfig.rootpath / 'docs' / 'configuration.rst'

    all_cfgs = []

    lines = []
    add = False
    indent = 0

    for line in file.read_text().splitlines():
        line = line
        stripped = line.strip()

        if add:
            if not indent and stripped:
                while line[indent] == ' ':
                    indent += 1

            if stripped and line[0] != ' ':
                all_cfgs.append(lines)
                add = False
                continue

            lines.append(line[indent:])

        if stripped.startswith('.. code-block:: yaml') or stripped.startswith('.. code-block:: yml'):
            add = True
            lines = []

    if add:
        all_cfgs.append(lines)

    for cfg_lines in all_cfgs:
        sample_cfg = '\n'.join(cfg_lines)

        map = yaml.yaml_rt.load(sample_cfg)
        CONFIG.load_config_dict(map)


def test_config_documentation_complete(pytestconfig):
    cfg_docs: Path = pytestconfig.rootpath / 'docs' / 'configuration.rst'
    cfg_model_dir: Path = pytestconfig.rootpath / 'src' / 'HABApp' / 'config' / 'models'
    assert cfg_model_dir.is_dir()

    documented_objs = set()

    # documented config
    current_module = ''
    for line in map(lambda x: x.strip().replace('  ', ''), cfg_docs.read_text().splitlines()):  # type: str
        if line.startswith('.. py:currentmodule::'):
            current_module = line[21:].strip()
            continue

        if line.startswith('.. autopydantic_model::'):
            obj_name = line[23:].strip()
            if current_module:
                obj_name = f'{current_module}.{obj_name}'
            assert obj_name not in documented_objs
            documented_objs.add(obj_name)

    # config implementation
    existing_objs = set()
    for module_name in [f.stem for f in cfg_model_dir.glob('**/*.py')]:
        module = getattr(HABApp.config.models, module_name)
        cfg_objs = [x[1] for x in getmembers(module, lambda x: isclass(x) and issubclass(x, BaseModel))]
        cfg_names = {
            f'{obj.__module__}.{obj.__qualname__}' for obj in cfg_objs if not obj.__module__.startswith('easyconfig.')
        }
        existing_objs.update(cfg_names)

        # we check this here to get the module with the error message
        missing = cfg_names - documented_objs
        assert not missing, module.__name__

    # ensure that everything that is implemented is documented
    assert existing_objs == documented_objs



from typing import Optional

from msgspec import Struct


# https://github.com/openhab/openhab-core/blob/main/bundles/org.openhab.core.io.rest/src/main/java/org/openhab/core/io/rest/internal/resources/beans/SystemInfoBean.java


class SystemInfoResp(Struct, rename='camel', kw_only=True):
    config_folder: str
    userdata_folder: str
    log_folder: Optional[str] = None
    java_version: Optional[str] = None
    java_vendor: Optional[str] = None
    java_vendor_version: Optional[str] = None
    os_name: Optional[str] = None
    os_version: Optional[str] = None
    os_architecture: Optional[str] = None
    available_processors: int
    free_memory: int
    total_memory: int
    start_level: int


class SystemInfoRootResp(Struct, rename='camel'):
    system_info: SystemInfoResp



from .base_event import OpenhabEvent
from .item_events import ItemStateEvent, ItemStateUpdatedEvent, ItemStateChangedEvent, ItemCommandEvent, ItemAddedEvent,\
    ItemUpdatedEvent, ItemRemovedEvent, ItemStatePredictedEvent, GroupStateUpdatedEvent, GroupStateChangedEvent
from .channel_events import ChannelTriggeredEvent, ChannelDescriptionChangedEvent
from .thing_events import ThingStatusInfoChangedEvent, ThingStatusInfoEvent, \
    ThingFirmwareStatusInfoEvent, ThingAddedEvent, ThingRemovedEvent, ThingUpdatedEvent, ThingConfigStatusInfoEvent
from .event_filters import ItemStateUpdatedEventFilter, ItemStateEventFilter, ItemStateChangedEventFilter, \
    ItemCommandEventFilter



from typing import FrozenSet, Mapping, Optional, Tuple, TYPE_CHECKING

from HABApp.openhab.definitions.values import PointValue
from HABApp.openhab.items.base_item import OpenhabItem, MetaData

if TYPE_CHECKING:
    Tuple = Tuple
    Optional = Optional
    FrozenSet = FrozenSet
    Mapping = Mapping
    MetaData = MetaData


class CallItem(OpenhabItem):
    """CallItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Tuple[str, ...] value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        return tuple(state.split(','))

    def set_value(self, new_value) -> bool:
        if isinstance(new_value, str):
            return super().set_value(tuple(new_value.split(',')))

        if isinstance(new_value, tuple):
            return super().set_value(new_value)

        assert new_value is None
        return super().set_value(tuple())


class LocationItem(OpenhabItem):
    """LocationItem which accepts and converts the data types from OpenHAB

    :ivar str name: |oh_item_desc_name|
    :ivar Optional[Tuple[float, float, Optional[float]]] value: |oh_item_desc_value|

    :ivar Optional[str] label: |oh_item_desc_label|
    :ivar FrozenSet[str] tags: |oh_item_desc_tags|
    :ivar FrozenSet[str] groups: |oh_item_desc_group|
    :ivar Mapping[str, MetaData] metadata: |oh_item_desc_metadata|
    """

    @staticmethod
    def _state_from_oh_str(state: str):
        value = tuple(state.split(','))
        assert 2 <= len(value) <= 3
        return value

    def set_value(self, new_value) -> bool:
        if isinstance(new_value, PointValue):
            return super().set_value(new_value.value)

        if new_value is None:
            return super().set_value(new_value)

        if isinstance(new_value, tuple):
            assert 2 <= len(new_value) <= 3
            return super().set_value(new_value)

        parts = new_value.split(',')    # type: list[Optional[str]]
        if len(parts) == 2:
            parts.append(None)
        assert len(parts) == 3
        return super().set_value(tuple(parts))
